from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from RPA.RobotLogListener import RobotLogListener

try:
    BuiltIn().import_library("RPA.RobotLogListener")
except RobotNotRunningError:
    pass


class CustomLibrary:
    SPECIAL_VALUE = 123456

    def __init__(self):
        listener = RobotLogListener()
        listener.register_protected_keywords(["CustomLibrary.special_keyword"])

    def special_keyword(self):
        print("will not be written to log")
        return self.SPECIAL_VALUE

    def assert_library_special_value(self, value):
        assert (
            value == self.SPECIAL_VALUE
        ), f"Value {value} did not match expected {self.SPECIAL_VALUE}"



import base64
import binascii
import collections
import copy
import json
import logging
import os
import traceback
from abc import abstractmethod, ABCMeta
from typing import Tuple

import requests
import yaml
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError

from RPA.core.helpers import import_by_name, required_env
from .utils import url_join, resolve_path


class RobocorpVaultError(RuntimeError):
    """Raised when there's problem with reading from Robocorp Vault."""


class Secret(collections.abc.Mapping):
    """Container for a secret with name, description, and
    multiple key-value pairs. Immutable and avoids logging
    internal values when possible.

    :param name:        Name of secret
    :param description: Human-friendly description for secret
    :param values:      Dictionary of key-value pairs stored in secret
    """

    def __init__(self, name, description, values):
        self._name = name
        self._desc = description
        self._dict = collections.OrderedDict(**values)

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._desc

    def update(self, kvpairs):
        self._dict.update(kvpairs)

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __contains__(self, key):
        return key in self._dict

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return "Secret(name={name}, keys=[{keys}])".format(
            name=self.name, keys=", ".join(str(key) for key in self.keys())
        )


class BaseSecretManager(metaclass=ABCMeta):
    """Abstract class for secrets management. Should be used as a
    base-class for any adapter implementation.
    """

    @abstractmethod
    def get_secret(self, secret_name):
        """Return ``Secret`` object with given name."""

    @abstractmethod
    def set_secret(self, secret: Secret):
        """Set a secret with a new value."""


class FileSecrets(BaseSecretManager):
    """Adapter for secrets stored in a database file. Supports only
    plaintext secrets, and should be used mainly for debugging.

    The path to the secrets file can be set with the
    environment variable ``RPA_SECRET_FILE``, or as
    an argument to the library.

    The format of the secrets file should be one of the following:

    .. code-block:: JSON

      {
        "name1": {
          "key1": "value1",
          "key2": "value2"
        },
        "name2": {
          "key1": "value1"
        }
      }

    OR

    .. code-block:: YAML

      name1:
        key1: value1
        key2: value2
      name2:
        key1: value1
    """

    SERIALIZERS = {
        ".json": (json.load, json.dump),
        ".yaml": (yaml.full_load, yaml.dump),
    }

    def __init__(self, secret_file="secrets.json"):
        self.logger = logging.getLogger(__name__)

        path = required_env("RPA_SECRET_FILE", secret_file)
        self.logger.info("Resolving path: %s", path)
        self.path = resolve_path(path)

        extension = self.path.suffix
        serializer = self.SERIALIZERS.get(extension)
        # NOTE(cmin764): This will raise instead of returning an empty secrets object
        #  because it is wrong starting from the "env.json" configuration level.
        if not serializer:
            raise ValueError(
                f"Not supported local vault secrets file extension {extension!r}"
            )
        self._loader, self._dumper = serializer

        self.data = self.load()

    def load(self):
        """Load secrets file."""
        try:
            with open(self.path, encoding="utf-8") as fd:
                data = self._loader(fd)

            if not isinstance(data, dict):
                raise ValueError("Invalid content format")

            return data
        except (IOError, ValueError) as err:
            self.logger.error("Failed to load secrets file: %s", err)
            return {}

    def save(self):
        """Save the secrets content to disk."""
        try:
            with open(self.path, "w", encoding="utf-8") as f:
                if not isinstance(self.data, dict):
                    raise ValueError("Invalid content format")
                self._dumper(self.data, f, indent=4)
        except (IOError, ValueError) as err:
            self.logger.error("Failed to save secrets file: %s", err)

    def get_secret(self, secret_name):
        """Get secret defined with given name from file.

        :param secret_name: Name of secret to fetch
        :returns:           Secret object
        :raises KeyError:   No secret with given name
        """
        values = self.data.get(secret_name)
        if values is None:
            raise KeyError(f"Undefined secret: {secret_name}")

        return Secret(secret_name, "", values)

    def set_secret(self, secret: Secret) -> None:
        """Set the secret value in the local Vault
        with the given ``Secret`` object.

        :param secret:                 A ``Secret`` object.
        :raises IOError, ValueError:   Writing the local vault failed.
        """
        self.data[secret.name] = dict(secret)
        self.save()


class RobocorpVault(BaseSecretManager):
    """Adapter for secrets stored in Robocorp Vault.

    The following environment variables should exist:

    - RC_API_SECRET_HOST:   URL to Robocorp Secrets API
    - RC_API_SECRET_TOKEN:  API token with access to Robocorp Secrets API
    - RC_WORKSPACE_ID:      Robocorp Workspace ID

    If the robot run is started from the Robocorp Control Room these environment
    variables will be configured automatically.
    """

    ENCRYPTION_SCHEME = "robocloud-vault-transit-v2"

    def __init__(self, *args, **kwargs):
        # pylint: disable=unused-argument
        self.logger = logging.getLogger(__name__)
        # Environment variables set by runner
        self._host = required_env("RC_API_SECRET_HOST")
        self._token = required_env("RC_API_SECRET_TOKEN")
        self._workspace = required_env("RC_WORKSPACE_ID")
        # Generated lazily on request
        self.__private_key = None
        self.__public_bytes = None

    @property
    def headers(self):
        """Default request headers."""
        return {"Authorization": f"Bearer {self._token}"}

    @property
    def params(self):
        """Default request parameters."""
        return {
            "encryptionScheme": self.ENCRYPTION_SCHEME,
            "publicKey": self._public_bytes,
        }

    @property
    def _private_key(self):
        """Cryptography private key object."""
        if self.__private_key is None:
            self.__private_key = rsa.generate_private_key(
                public_exponent=65537, key_size=4096, backend=default_backend()
            )

        return self.__private_key

    @property
    def _public_bytes(self):
        """Serialized public key bytes."""
        if self.__public_bytes is None:
            self.__public_bytes = base64.b64encode(
                self._private_key.public_key().public_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PublicFormat.SubjectPublicKeyInfo,
                )
            )

        return self.__public_bytes

    def create_secret_url(self, name):
        """Create a URL for a specific secret."""
        return url_join(
            self._host, "secrets-v1", "workspaces", self._workspace, "secrets", name
        )

    def create_public_key_url(self):
        """Create a URL for encryption public key."""
        return url_join(
            self._host,
            "secrets-v1",
            "workspaces",
            self._workspace,
            "secrets",
            "publicKey",
        )

    def get_secret(self, secret_name):
        """Get secret defined with given name from Robocorp Vault.

        :param secret_name:         Name of secret to fetch
        :returns:                   Secret object
        :raises RobocorpVaultError: Error with API request or response payload
        """
        url = self.create_secret_url(secret_name)

        try:
            response = requests.get(url, headers=self.headers, params=self.params)
            response.raise_for_status()

            payload = response.json()
            payload = self._decrypt_payload(payload)
        except InvalidTag as e:
            self.logger.debug(traceback.format_exc())
            raise RobocorpVaultError("Failed to validate authentication tag") from e
        except Exception as exc:
            self.logger.debug(traceback.format_exc())
            raise RobocorpVaultError from exc

        return Secret(payload["name"], payload["description"], payload["values"])

    def _decrypt_payload(self, payload):
        payload = copy.deepcopy(payload)

        fields = payload.pop("encryption", None)
        if fields is None:
            raise KeyError("Missing encryption fields from response")

        scheme = fields["encryptionScheme"]
        if scheme != self.ENCRYPTION_SCHEME:
            raise ValueError(f"Unexpected encryption scheme: {scheme}")

        aes_enc = base64.b64decode(fields["encryptedAES"])
        aes_tag = base64.b64decode(fields["authTag"])
        aes_iv = base64.b64decode(fields["iv"])

        # Decrypt AES key using our private key
        aes_key = self._private_key.decrypt(
            aes_enc,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

        # Decrypt actual value using decrypted AES key
        ciphertext = base64.b64decode(payload.pop("value")) + aes_tag
        data = AESGCM(aes_key).decrypt(binascii.hexlify(aes_iv), ciphertext, b"")
        payload["values"] = json.loads(data)

        return payload

    def set_secret(self, secret: Secret) -> None:
        """Set the secret value in the Vault. Note that the secret possibly
        consists of multiple key-value pairs, which will all be overwritten
        with the values given here. So don't try to update only one item
        of the secret, update all of them.

        :param secret: A ``Secret`` object
        """
        value, aes_iv, aes_key, aes_tag = self._encrypt_secret_value_with_aes(secret)
        pub_key = self.get_publickey()
        aes_enc = self._encrypt_aes_key_with_public_rsa(aes_key, pub_key)

        payload = {
            "description": secret.description,
            "encryption": {
                "authTag": aes_tag.decode(),
                "encryptedAES": aes_enc.decode(),
                "encryptionScheme": self.ENCRYPTION_SCHEME,
                "iv": aes_iv.decode(),
            },
            "name": secret.name,
            "value": value.decode(),
        }

        url = self.create_secret_url(secret.name)
        try:
            response = requests.put(url, headers=self.headers, json=payload)
            response.raise_for_status()
        except Exception as e:
            self.logger.debug(traceback.format_exc())
            if response.status_code == 403:
                raise RobocorpVaultError(
                    "Failed to set secret value. Does your token have write access?"
                ) from e
            raise RobocorpVaultError("Failed to set secret value.") from e

    def get_publickey(self) -> bytes:
        """Get the public key for AES encryption with the existing token."""
        url = self.create_public_key_url()
        try:
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
        except Exception as e:
            self.logger.debug(traceback.format_exc())
            raise RobocorpVaultError(
                "Failed to fetch public key. Is your token valid?"
            ) from e

        return response.content

    @staticmethod
    def _encrypt_secret_value_with_aes(
        secret: Secret,
    ) -> Tuple[bytes, bytes, bytes, bytes]:
        def generate_aes_key() -> Tuple[bytes, bytes]:
            aes_key = AESGCM.generate_key(bit_length=256)
            aes_iv = os.urandom(16)
            return aes_key, aes_iv

        def split_auth_tag_from_encrypted_value(
            encrypted_value: bytes,
        ) -> Tuple[bytes, bytes]:
            """AES auth tag is the last 16 bytes of the AES encrypted value.
            Split the tag from the value, as that is required for the API.
            """
            aes_tag = encrypted_value[-16:]
            trimmed_encrypted_value = encrypted_value[:-16]
            return trimmed_encrypted_value, aes_tag

        value = json.dumps(dict(secret)).encode()

        aes_key, aes_iv = generate_aes_key()
        encrypted_value = AESGCM(aes_key).encrypt(aes_iv, value, b"")
        encrypted_value, aes_tag = split_auth_tag_from_encrypted_value(encrypted_value)

        return (
            base64.b64encode(encrypted_value),
            base64.b64encode(aes_iv),
            aes_key,
            base64.b64encode(aes_tag),
        )

    @staticmethod
    def _encrypt_aes_key_with_public_rsa(aes_key: bytes, public_rsa: bytes) -> bytes:
        pub_decoded = base64.b64decode(public_rsa)
        public_key = serialization.load_der_public_key(pub_decoded)

        aes_enc = public_key.encrypt(
            aes_key,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

        return base64.b64encode(aes_enc)


class Vault:
    """`Vault` is a library for interacting with secrets stored in the `Robocorp
    Control Room Vault`_ (by default) or file-based secrets, which can be taken
    into use by setting some environment variables.

    Robocorp Vault relies on environment variables, which are normally set
    automatically by the Robocorp Work Agent or Assistant when a run is
    initialized by the Robocorp Control Room. When developing robots locally
    in VSCode, you can use the `Robocorp Code Extension`_ to set these
    variables automatically as well.

    Alternatively, you may set these environment variable manually using
    `rcc`_ or directly in some other fashion. The specific variables which
    must exist are:

    - ``RC_API_SECRET_HOST``: URL to Robocorp Vault API
    - ``RC_API_SECRET_TOKEN``: API Token for Robocorp Vault API
    - ``RC_WORKSPACE_ID``: Control Room Workspace ID

    .. _Robocorp Control Room Vault: https://robocorp.com/docs/development-guide/variables-and-secrets/vault
    .. _Robocorp Code Extension: https://robocorp.com/docs/developer-tools/visual-studio-code/extension-features#connecting-to-control-room-vault
    .. _rcc: https://robocorp.com/docs/rcc/workflow

    File-based secrets can be set by defining two environment variables.

    - ``RPA_SECRET_MANAGER``: RPA.Robocorp.Vault.FileSecrets
    - ``RPA_SECRET_FILE``: Absolute path to the secrets database file

    Example content of local secrets file:

    .. code-block:: json

        {
            "swaglabs": {
                "username": "standard_user",
                "password": "secret_sauce"
            }
        }

    OR

    .. code-block:: YAML

        swaglabs:
            username: standard_user
            password: secret_sauce

    **Examples of Using Secrets in a Robot**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library    Collections
        Library    RPA.Robocorp.Vault

        *** Tasks ***
        Reading secrets
            ${secret}=    Get Secret  swaglabs
            Log Many      ${secret}

        Modifying secrets
            ${secret}=          Get Secret      swaglabs
            ${level}=           Set Log Level   NONE
            Set To Dictionary   ${secret}       username    nobody
            Set Log Level       ${level}
            Set Secret          ${secret}


    **Python**

    .. code-block:: python

        from RPA.Robocorp.Vault import Vault

        VAULT = Vault()

        def reading_secrets():
            print(f"My secrets: {VAULT.get_secret('swaglabs')}")

        def modifying_secrets():
            secret = VAULT.get_secret("swaglabs")
            secret["username"] = "nobody"
            VAULT.set_secret(secret)

    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self, *args, **kwargs):
        """The selected adapter can be set with the environment variable
        ``RPA_SECRET_MANAGER``, or the keyword argument ``default_adapter``.
        Defaults to Robocorp Vault if not defined.

        All other library arguments are passed to the adapter.

        :param default_adapter: Override default secret adapter
        """
        self.logger = logging.getLogger(__name__)

        default = kwargs.pop("default_adapter", RobocorpVault)
        adapter = required_env("RPA_SECRET_MANAGER", default)

        self._adapter_factory = self._create_factory(adapter, args, kwargs)
        self._adapter = None

        try:
            BuiltIn().import_library("RPA.RobotLogListener")
        except RobotNotRunningError:
            pass

    @property
    def adapter(self):
        if self._adapter is None:
            self._adapter = self._adapter_factory()

        return self._adapter

    def _create_factory(self, adapter, args, kwargs):
        if isinstance(adapter, str):
            adapter = import_by_name(adapter, __name__)

        if not issubclass(adapter, BaseSecretManager):
            raise ValueError(
                f"Adapter '{adapter}' does not inherit from BaseSecretManager"
            )

        def factory():
            return adapter(*args, **kwargs)

        return factory

    def get_secret(self, secret_name: str) -> Secret:
        """Read a secret from the configured source, e.g. Robocorp Vault,
        and return it as a ``Secret`` object.

        :param secret_name: Name of secret
        """
        return self.adapter.get_secret(secret_name)

    def set_secret(self, secret: Secret) -> None:
        """Overwrite an existing secret with new values.

        Note: Only allows modifying existing secrets, and replaces
              all values contained within it.

        :param secret: Secret as a ``Secret`` object, from e.g. ``Get Secret``
        """
        self.adapter.set_secret(secret)



import pytest
from RPA.JSON import JSON


ORDERS = """
{
  "clients": [
    {
      "name": "Johnny Example",
      "email": "john@example.com",
      "orders": [
          {"address": "Streetroad 123", "price": 103.20},
          {"address": "Streetroad 123", "price": 98.99}
      ]
    },
    {
      "name": "Jane Example",
      "email": "jane@example.com",
      "orders": [
          {"address": "Waypath 321", "price": 22.00},
          {"address": "Streetroad 123", "price": 2330.01}
      ]
    }
  ]
}
"""

MARKS = [
    '{"People": [{"Name": "Mark", "Email": "mark@robocorp.com"}, {"Name": "Jane", '
    '"Extra": 1}]}',
    '{"People": {"a": 1, "b": {"Name": "Mark", "Email": "mark@robocorp.com"}, "c": '
    '{"Name": "Jane", "Extra": 1}}}',
    '{"People": {"a": 1, "b": {"z": {"Name": "Mark", "Email": "mark@robocorp.com"}}, '
    '"c": {"Name": "Jane", "Extra": 1}}}',
]


@pytest.fixture
def lib():
    return JSON()


def str2json(json_string, *, lib):
    return lib.convert_string_to_json(json_string.strip())


@pytest.fixture
def orders(lib):
    return str2json(ORDERS, lib=lib)


@pytest.fixture(params=MARKS)
def mark(lib, request):
    return str2json(request.param, lib=lib)


def test_json_expression(lib, orders):
    result = lib.get_values_from_json(orders, "$..address")
    assert result == [
        "Streetroad 123",
        "Streetroad 123",
        "Waypath 321",
        "Streetroad 123",
    ]


def test_json_expression_filter(lib, orders):
    result = lib.get_values_from_json(orders, "$..orders[?(@.price>100)]")
    assert result == [
        {"address": "Streetroad 123", "price": 103.2},
        {"address": "Streetroad 123", "price": 2330.01},
    ]


def test_json_single_value(lib, orders):
    assert lib.get_value_from_json(orders, "$.clients[0].name") == "Johnny Example"


def test_json_single_value_default(lib, orders):
    assert (
        lib.get_value_from_json(orders, "$.clients[3].name", default="Notexist")
        == "Notexist"
    )


def test_json_update(lib, orders):
    result = lib.update_value_to_json(orders, "$.clients[*].name", "John Malkovich")
    assert all(client["name"] == "John Malkovich" for client in result["clients"])


@pytest.mark.parametrize(
    "expr",
    [
        '$.People[?(@..Name=="Mark")]',
        '$.People[?(@..Name=="Mark") & (@..Email=="mark@robocorp.com")]',
        '$[?(@..Name=="Other")] | $[?(@..Email=="mark@robocorp.com")]',
        '($[?(@..Name=="Other")] | $[?(@..Email=="mark@robocorp.com")])[?(@..Email=="mark@robocorp.com")]',
        '($|($.*)|($.*.*))[?(@.Name=="Mark")]',
    ],
)
def test_delete_from_json(lib, mark, expr):
    mark_dict = lib.delete_from_json(mark, expr)
    mark_str = lib.convert_json_to_string(mark_dict)
    assert "Mark" not in mark_str



import time
from typing import Optional
from PIL import ImageOps

from RPA.core.geometry import Region
from RPA.Desktop.keywords import LibraryContext, keyword, screen, HAS_RECOGNITION

if HAS_RECOGNITION:
    from RPA.recognition import ocr  # pylint: disable=no-name-in-module


def ensure_recognition():
    if not HAS_RECOGNITION:
        raise ValueError(
            "Keyword requires OCR features, please install the "
            "rpaframework-recognition package"
        )


class TextKeywords(LibraryContext):
    """Keywords for reading screen information and content."""

    @keyword
    def read_text(self, locator: Optional[str] = None, invert: bool = False):
        """Read text using OCR from the screen, or an area of the
        screen defined by the given locator.

        :param locator: Location of element to read text from
        :param invert:  Invert image colors, useful for reading white text
                        on dark background

        Usage examples:

        .. code-block:: robotframework

            ${label_region}=  Find Element  image:label.png
            ${value_region}=  Move Region   ${label_region}  100  0
            ${text}=          Read Text     ${value_region}

        .. code-block:: python

            label_region = desktop.find_element("image:label.png")
            value_region = desktop.move_region(label_region, 100, 0)
            text = desktop.read_text(value_region)

        """
        ensure_recognition()

        if locator is not None:
            element = self.ctx.wait_for_element(locator)
            if not isinstance(element, Region):
                raise ValueError("Locator must resolve to a region")

            self.logger.info("Reading text from element: %s", element)
            image = screen.grab(element)
        else:
            self.logger.info("Reading text from screen")
            image = screen.grab()

        screen.log_image(image)

        if invert:
            image = ImageOps.invert(image)

        start_time = time.time()
        text = ocr.read(image)
        self.logger.info("Read text in %.2f seconds", time.time() - start_time)

        return text



# pylint: disable=missing-class-docstring
import logging
import platform
import time
from typing import Optional
from comtypes import COMError


if platform.system() == "Windows":
    from SapGuiLibrary import SapGuiLibrary
else:
    logging.getLogger(__name__).warning(
        "RPA.SapGuiLibrary library works only on Windows platform"
    )

    class SapGuiLibrary:
        """Keywords are only available in Windows."""

        def __init__(self, *args, **kwargs):
            del args, kwargs


class SAP(SapGuiLibrary):
    __doc__ = (
        "This library wraps the upstream "
        "[https://frankvanderkuur.github.io/SapGuiLibrary.html|SapGuiLibrary]."
        "\n\n" + SapGuiLibrary.__doc__
    )

    ROBOT_LIBRARY_SCOPE = "GLOBAL"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)

        if platform.system() != "Windows":
            self.logger.warning("SAP requires Windows dependencies to work.")

    def get_statusbar_type(self, window: Optional[str] = "wnd[0]") -> str:
        """Retrieves the messageType in the statusbar the given window.

        Takes screenshot on error.

        :param window: locator for the statusbar, default 'wnd[0]'
        :return: messageType of the /sbar element or empty string
        """
        return_value = ""
        try:
            return_value = self.session.findById(window + "/sbar").messageType
        except COMError as ex:
            self.take_screenshot()
            raise ValueError("Cannot find window with locator '%s'" % window) from ex
        return return_value

    def focus_and_click(self, element_id: str, wait_time: Optional[float] = None):
        """Set focus into the element and click it.

        Note. The default library wait time can be adjusted using `Set Explicit Wait`
        keyword. Library's `explicit_wait` will be used if `wait_time` parameter
        is not set.

        :param element_id: locator for the element
        :param wait_time: the wait time after the action
        """
        sleeptime = wait_time or self.explicit_wait
        self.set_focus(element_id)
        self.click_element(element_id)
        if sleeptime and sleeptime > 0:
            time.sleep(sleeptime)

    def focus_and_input_text(
        self, element_id: str, text: str, wait_time: Optional[float] = None
    ):
        """Set focus into the element and input text into it.

        Note. The default library wait time can be adjusted using `Set Explicit Wait`
        keyword. Library's `explicit_wait` will be used if `wait_time` parameter
        is not set.

        :param element_id: locator for the element
        :param text: text to be inputted
        :param wait_time: the wait time after the action
        """
        sleeptime = wait_time or self.explicit_wait
        self.set_focus(element_id)
        self.input_text(element_id, text)
        if sleeptime and sleeptime > 0:
            time.sleep(sleeptime)

    def generic_input_password(self, element_id, password):
        """Inserts the given password into the text field identified by locator.

        This keyword does NOT depend on the type of the element. Will set
        password to the element's 'text' attribute if possible.

        The password is not recorded in the log.

        :param element_id: locator for the element
        :param password: password to be inputted
        """
        self._generic_input(element_id, password, False)

    def generic_input_text(self, element_id, text):
        """Inserts the given text into the text field identified by locator.
        Use keyword `input password` to insert a password in a text field.

        This keyword does NOT depend on the type of the element. Will set
        text to the element's 'text' attribute if possible.

        :param element_id: locator for the element
        :param text: text to be inputted
        """
        self._generic_input(element_id, text)

    def get_element_type_of_object(self, element):
        """Returns the Sap element type for the given element.

        :param element: SAP element
        :return: type of the SAP element
        """
        try:
            return element.type
        except COMError as ex:
            self.take_screenshot()
            message = "Element does not have 'type' attribute: '%s'" % type(element)
            raise ValueError(message) from ex

    def _generic_input(self, element_id, text, log_text=True):
        element = self.session.findById(element_id)
        element_type = self.get_element_type_of_object(element)
        if hasattr(element, "text"):
            element.text = text
            self.session.findById(element_id).text = text
            if log_text:
                self.logger.info(
                    "Typing text '%s' into text field '%s'." % (text, element_id)
                )
            else:
                self.logger.info("Typing password into text field '%s'." % element_id)
            time.sleep(self.explicit_wait)
        else:
            self.take_screenshot()
            message = "Cannot set text for element type '%s'" % element_type
            raise ValueError(message)

    def generic_click_element(self, element_id, click_type="press"):
        """Performs a single click on a given element.

        Differs from `Click Element` keyword so that element type is ignored, instead
        `click_type` (either `press` or `select`) is performed on the element if
        possible.

        In case you want to change a value of an element like checkboxes of selecting
        an option in dropdown lists, use `select checkbox` or
        `select from list by label` instead.

        :param element_id: locator for the element
        :param click_type: either ``press`` (default) or ``select``
        """
        element = self.session.findById(element_id)
        element_type = self.get_element_type_of_object(element)
        if click_type.lower() == "select" and hasattr(element, "select"):
            element.select()
        if click_type.lower() == "press" and hasattr(element, "press"):
            element.press()
        else:
            self.take_screenshot()
            message = (
                "You cannot use '%s' on element type '%s', maybe use "
                "'select checkbox' instead?" % (click_type, element_type)
            )
            raise Warning(message)
        time.sleep(self.explicit_wait)

    def get_cell_value(self, table_id, row_num, col_id):
        """Returns the cell value for the specified cell.

        :param table_id: locator for the table element
        :param row_num: table row number
        :param col_id: table cell id
        :return: text in the specified cell
        """
        self.element_should_be_present(table_id)

        try:
            cellValue = self.session.findById(table_id).getCell(row_num, col_id).text
            return cellValue
        except COMError as ex:
            self.take_screenshot()
            message = "Cannot find Column_id '%s'." % col_id
            raise ValueError(message) from ex



import base64
import contextlib
import os
import tempfile
from contextlib import contextmanager

import mock
import pytest
from RPA.Crypto import Crypto, Hash
from RPA.Robocorp.Vault import Secret


@contextmanager
def temp_path(content=None, suffix=None):
    with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as temp:
        path = temp.name
        if content:
            temp.write(content)
    try:
        yield path
    finally:
        with contextlib.suppress(FileNotFoundError):
            os.unlink(path)


def test_generate_key():
    key = Crypto().generate_key()
    assert isinstance(key, str)
    assert len(key) > 0

    keys = [Crypto().generate_key() for _ in range(50)]
    assert len(set(keys)) == len(keys)


def test_set_key():
    lib = Crypto()
    assert lib._key is None

    key = lib.generate_key()
    assert lib._key is None

    lib.use_encryption_key(key)
    assert lib._key is not None

    with pytest.raises(ValueError):
        lib.use_encryption_key("somethingelse")


def test_hash_string():
    result = Crypto().hash_string("avalue")
    assert result == "eNOTzEOqI0k2ijVhEmKow1YPFGo="


def test_hash_string_unicode():
    result = Crypto().hash_string("谷歌翻译是你的朋友")
    assert result == "J5TSpa76g8qV1pws6H6ztDEfYeo="


def test_hash_string_method():
    result = Crypto().hash_string("avalue", method=Hash.MD5)
    assert result == "3ObaKiohbtSuG4Q8+kIKTg=="


def test_hash_file():
    with temp_path("avalue".encode("utf-8")) as infile:
        result = Crypto().hash_file(infile)
        assert result == "eNOTzEOqI0k2ijVhEmKow1YPFGo="


def test_encrypt_decrypt_string():
    lib = Crypto()

    key = lib.generate_key()
    lib.use_encryption_key(key)

    text = "An example string\nWith some secret content"
    encrypted = lib.encrypt_string(text)

    assert encrypted != text
    assert base64.urlsafe_b64decode(encrypted) != text

    decrypted = lib.decrypt_string(encrypted)
    assert decrypted == text


def test_decrypt_wrong_key():
    lib = Crypto()

    key1 = lib.generate_key()
    key2 = lib.generate_key()
    assert key1 != key2

    lib.use_encryption_key(key1)
    text = "An example string\nWith some secret content"
    token = lib.encrypt_string(text)

    lib.use_encryption_key(key2)
    with pytest.raises(ValueError):
        lib.decrypt_string(token)


def test_encrypt_decrypt_file():
    lib = Crypto()

    key = lib.generate_key()
    lib.use_encryption_key(key)

    text = "An example string\nWith some secret content"

    with temp_path() as encrypted:
        with temp_path(text.encode("utf-8")) as original:
            result = lib.encrypt_file(original, encrypted)
            assert result == encrypted
            assert os.stat(result).st_size > 0

        with temp_path() as decrypted:
            result = lib.decrypt_file(encrypted, decrypted)
            assert result == decrypted
            with open(result) as resultfile:
                assert resultfile.read() == text


def test_encrypt_file_default_paths():
    lib = Crypto()

    key = lib.generate_key()
    lib.use_encryption_key(key)

    text = "An example string\nWith some secret content"

    with temp_path(text.encode("utf-8")) as original:
        encrypted, decrypted = "", ""
        try:
            encrypted = lib.encrypt_file(original)
            assert encrypted == original + ".enc"

            decrypted = lib.decrypt_file(encrypted)
            assert decrypted == original
        finally:
            with contextlib.suppress(FileNotFoundError):
                os.unlink(encrypted)
                os.unlink(decrypted)


def test_encrypt_file_suffix():
    lib = Crypto()

    key = lib.generate_key()
    lib.use_encryption_key(key)

    text = "An example string\nWith some secret content"

    with temp_path(text.encode("utf-8")) as original:
        with temp_path(suffix=".bin") as encrypted:
            lib.encrypt_file(original, encrypted)
            result = lib.decrypt_file(encrypted)
            try:
                assert os.path.isfile(result)
                assert result.endswith(".dec.bin")
            finally:
                with contextlib.suppress(FileNotFoundError):
                    os.unlink(result)


def test_set_key_vault_no_key():
    lib = Crypto()
    lib._vault = mock_vault = mock.Mock()

    key = lib.generate_key()
    mock_vault.get_secret.return_value = Secret("MockSecret", "", {"key": key})

    lib.use_encryption_key_from_vault("SomeKeyValue")
    assert mock_vault.get_secret.called_once_with("SomeKeyValue")
    assert lib._key is not None


def test_set_key_vault_key():
    lib = Crypto()
    lib._vault = mock_vault = mock.Mock()

    key = lib.generate_key()
    mock_vault.get_secret.return_value = Secret(
        "MockSecret", "", {"first": "something", "second": key}
    )

    lib.use_encryption_key_from_vault("SomeKeyValue", "second")
    assert mock_vault.get_secret.called_once_with("SomeKeyValue")
    assert lib._key is not None


def test_set_key_vault_error_multiple():
    lib = Crypto()
    lib._vault = mock_vault = mock.Mock()

    key = lib.generate_key()
    mock_vault.get_secret.return_value = Secret(
        "MockSecret", "", {"first": "something", "second": key}
    )

    with pytest.raises(ValueError):
        lib.use_encryption_key_from_vault("SomeKeyValue")
    assert mock_vault.get_secret.called_once_with("SomeKeyValue")
    assert lib._key is None


def test_set_key_vault_error_empty():
    lib = Crypto()
    lib._vault = mock_vault = mock.Mock()

    key = lib.generate_key()
    mock_vault.get_secret.return_value = Secret("MockSecret", "", {})

    with pytest.raises(ValueError):
        lib.use_encryption_key_from_vault("SomeKeyValue")
    assert mock_vault.get_secret.called_once_with("SomeKeyValue")
    assert lib._key is None



from dataclasses import dataclass
import datetime
import logging
from robot.libraries.BuiltIn import (
    BuiltIn,
    RobotNotRunningError,
)
import tweepy
from tweepy.error import TweepError

from RPA.core.helpers import required_env, required_param
from RPA.core.notebook import notebook_json
from RPA.RobotLogListener import RobotLogListener


try:
    BuiltIn().import_library("RPA.RobotLogListener")
except RobotNotRunningError:
    pass


@dataclass
class Tweet:
    """Represents Tweet"""

    created_at: datetime.datetime
    id: int
    tweet_id_str: str
    text: str
    in_reply_to_screen_name: str
    lang: str
    name: str
    screen_name: str
    hashtags: list
    is_truncated: bool = False
    favorite_count: int = 0
    retweeted: bool = False
    retweet_count: int = 0


class Twitter:
    """`Twitter` is a library for accessing Twitter using developer API.
    The library extends `tweepy`_ library.

    Authorization credentials can be given as parameters for ``authorize`` keyword
    or keyword can read them in as environment variables:

    - `TWITTER_CONSUMER_KEY`
    - `TWITTER_CONSUMER_SECRET`
    - `TWITTER_ACCESS_TOKEN`
    - `TWITTER_ACCESS_TOKEN_SECRET`

    Library usage requires Twitter developer credentials.
    Those can be requested from `Twitter developer site`_

    .. _tweepy:
        http://docs.tweepy.org/en/latest/index.html

    .. _Twitter developer site:
        https://developer.twitter.com/

    **Examples**

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Twitter

        *** Tasks ***
        Get user tweets and like them
            [Setup]   Authorize
            @{tweets}=   Get User Tweets   username=niinisto   count=5
            FOR   ${tweet}  IN   @{tweets}
                Like   ${tweet}
            END

    .. code-block:: python

        from RPA.Twitter import Twitter

        library = Twitter()
        library.authorize()
        tweets = library.get_user_tweets(username="niinisto", count=5)
        for tw in tweets:
            library.like(tw)
        tweets = library.text_search_tweets(query="corona trump")
        for tw in tweets:
            print(tw.text)
        user = library.get_user_profile("niinisto")
        library.follow(user)
        library.tweet("first tweet")
        me = library.get_me()
        print(me)
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self._auth = None
        self.api = None
        self._me = None
        listener = RobotLogListener()
        listener.register_protected_keywords("authorize")
        listener.only_info_level(
            [
                "get_me",
                "get_user_tweets",
                "text_search_tweets",
                "get_user_profile",
                "tweet",
                "like",
                "unlike",
            ]
        )

    def authorize(
        self,
        consumer_key: str = None,
        consumer_secret: str = None,
        access_token: str = None,
        access_token_secret: str = None,
    ) -> None:
        """Authorize to Twitter API

        :param consumer_key: app consumer key
        :param consumer_secret: app consumer secret
        :param access_token: user access token
        :param access_token_secret: user access token secret
        """
        if consumer_key is None:
            consumer_key = required_env("TWITTER_CONSUMER_KEY")
        if consumer_secret is None:
            consumer_secret = required_env("TWITTER_CONSUMER_SECRET")
        if access_token is None:
            access_token = required_env("TWITTER_ACCESS_TOKEN")
        if access_token_secret is None:
            access_token_secret = required_env("TWITTER_ACCESS_TOKEN_SECRET")
        self._auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
        self._auth.set_access_token(access_token, access_token_secret)
        self.api = tweepy.API(self._auth, wait_on_rate_limit=True)
        try:
            self.api.verify_credentials()
            self.logger.info("Twitter authentication success")
            self._me = self.api.me()
        except TweepError as e:
            self.logger.error("Error during Twitter authentication: %s", str(e))
            raise TweepError from e

    def get_me(self) -> dict:
        """Get Twitter profile of authenticated user

        :return: user profile as dictionary or `None`
        """
        data = self._me._json if self._me else None  # pylint: disable=W0212
        notebook_json(data)
        return data

    def get_user_tweets(self, username: str = None, count: int = 100) -> list:
        """Get user tweets

        :param username: whose tweets to get
        :param count: maximum number of tweets, defaults to 100
        :return: list of user tweets
        """
        required_param(username, "get_user_tweets")
        tweets = []
        try:
            # Pulling individual tweets from query
            for tweet in self.api.user_timeline(id=username, count=count):
                # Adding to list that contains all tweets
                tw = Tweet(
                    created_at=tweet.created_at,
                    id=tweet.id,
                    tweet_id_str=tweet.id_str,
                    text=tweet.text,
                    in_reply_to_screen_name=tweet.in_reply_to_screen_name,
                    lang=tweet.lang,
                    name=tweet.user.name,
                    screen_name=tweet.user.screen_name,
                    hashtags=[ht["text"] for ht in tweet.entities["hashtags"]],
                    is_truncated=tweet.truncated,
                    favorite_count=tweet.favorite_count,
                    retweeted=tweet.retweeted,
                    retweet_count=tweet.retweet_count,
                )
                tweets.append(tw)
        except TweepError as e:
            self.logger.warning("Twitter timeline failed: %s", str(e))
        return tweets

    def text_search_tweets(
        self,
        query: str = None,
        count: int = 100,
        geocode: str = None,
        lang: str = None,
        locale: str = None,
        result_type: str = "mixed",
        until: str = None,
        since_id: str = None,
        max_id: str = None,
    ) -> list:
        """Search tweets defined by search query

        Results types:

        - mixed : include both popular and real time results in the response
        - recent : return only the most recent results in the response
        - popular : return only the most popular results in the response

        :param query: search query string of 500 characters maximum,
            including operators
        :param count: maximum number of tweets, defaults to 100
        :param geocode: tweets by users located within a given
            radius of the given latitude/longitude
        :param lang: language code of tweets
        :param locale: language of the query you are sending
        :param result_type: type of search results you would prefer to receive,
            default "mixed"
        :param until: tweets created before the given date
        :param since_id: Returns only statuses with an ID greater than
        :param max_id: only statuses with an ID less than
        :return: list of matching tweets
        """
        required_param(query, "text_search_tweets")
        tweets = []
        try:
            # Pulling individual tweets from query
            for tweet in self.api.search(
                q=query,
                count=count,
                geocode=geocode,
                lang=lang,
                locale=locale,
                result_type=result_type,
                until=until,
                since_id=since_id,
                max_id=max_id,
            ):
                tw = Tweet(
                    created_at=tweet.created_at,
                    id=tweet.id,
                    tweet_id_str=tweet.id_str,
                    text=tweet.text,
                    in_reply_to_screen_name=tweet.in_reply_to_screen_name,
                    lang=tweet.lang,
                    name=tweet.user.name,
                    screen_name=tweet.user.screen_name,
                    hashtags=[ht["text"] for ht in tweet.entities["hashtags"]],
                    is_truncated=tweet.truncated,
                    favorite_count=tweet.favorite_count,
                    retweeted=tweet.retweeted,
                    retweet_count=tweet.retweet_count,
                )
                tweets.append(tw)
        except TweepError as e:
            self.logger.warning("Twitter search failed: %s", str(e))
        return tweets

    def get_user_profile(self, username: str = None) -> dict:
        """Get user's Twitter profile

        :param username: whose profile to get
        :return: profile as dictionary
        """
        required_param(username, "get_user_profile")
        try:
            profile = self.api.get_user(username)
            data = profile._json  # pylint: disable=W0212
            notebook_json(data)
            return data
        except TweepError:
            return None

    def tweet(self, content: str = None) -> None:
        """Make a tweet with content

        :param content: text for the status update
        """
        required_param(content, "tweet")
        self.api.update_status(content)

    def like(self, tweet: Tweet = None) -> bool:
        """Like a tweet

        :param tweet: as a class `Tweet`
        :return: `True` if Tweet was liked, `False` if not
        """
        required_param(tweet, "like")
        try:
            self.api.create_favorite(tweet.id)
            return True
        except TweepError:
            self.logger.warning(
                'Could not like tweet "%s" by user "%s"',
                tweet.text,
                tweet.screen_name,
            )
            return False

    def unlike(self, tweet: Tweet = None) -> bool:
        """Unlike a tweet

        :param tweet: as a class `Tweet`
        :return: `True` if Tweet was unliked, `False` if not
        """
        required_param(tweet, "unlike")
        try:
            self.api.destroy_favorite(tweet.id)
            return True
        except TweepError:
            self.logger.warning(
                'Could not unlike tweet "%s" by user "%s"',
                tweet.text,
                tweet.screen_name,
            )
            return False

    def follow(self, user: str = None) -> bool:
        """Follow Twitter user

        :param user: screen name of the user
        :return:  `True` if user was followed, `False` if not
        """
        required_param(user, "follow")
        try:
            self.api.create_friendship(user)
            return True
        except TweepError:
            self.logger.warning("Could not follow user: %s", user)
            return False

    def unfollow(self, user: str = None) -> bool:
        """Unfollow Twitter user

        :param user: screen name of the user
        :return:  `True` if user was followed, `False` if not
        """
        required_param(user, "unfollow")
        try:
            self.api.destroy_friendship(user)
            return True
        except TweepError:
            self.logger.warning("Could not unfollow user: %s", user)
            return False



import logging
from enum import Enum
from typing import List, Optional

from pyotp import HOTP, TOTP
from requests_oauthlib import OAuth2Session
from robot.api.deco import keyword

from RPA.Robocorp.Vault import Vault
from RPA.Robocorp.utils import protect_keywords


class OTPMode(Enum):
    """Enumeration for type of OTP to use."""

    TIME = "TIME"
    COUNTER = "COUNTER"


class TOTPNotSetError(Exception):
    """Raised when TOTP (Time-based One-Time Password) has not been set."""

    ERROR_MSG = (
        "TOTP (Time-based One-Time Password) can be set during the library "
        "initialization. With `Use MFA Secret From Vault` keyword or with "
        "`Set Time Based OTP` keyword."
    )


class HOTPNotSetError(Exception):
    """Raised when HOTP (HMAC One-Time Password) has not been set."""

    ERROR_MSG = (
        "HOTP (HMAC One-Time Password) can be set during the library "
        "initialization. With `Use MFA Secret From Vault` keyword or with "
        "`Set Counter Based OTP` keyword."
    )


class OAuth2NotSetError(Exception):
    """Raised when OAuth2 session object isn't initialized but used."""

    ERROR_MSG = (
        "OAuth2 session required but not initialized, please call the "
        "`Generate OAuth URL` keyword first."
    )


class MFA:
    """**RPA.MFA** is a library intended mainly for generating one-time passwords (OTP)
    and not only, as OAuth2 support was introduced lately.

    Available with **rpaframework** version: 19.4.0

    Based on the `pyotp <https://pypi.org/project/pyotp/>`_ and
    `requests_oauthlib <https://pypi.org/project/requests-oauthlib/>`_ packages. It
    provides support for both MFA with the ``* OTP`` related keywords and OAuth2
    "Authorization Code Flow" with the ``* OAuth *`` related keywords.

    In the below example the **mfa** secret we are reading from the Robocorp
    Vault is the passcode generated by the Authenticator service. The passcode
    value is stored into the Vault with key **otpsecret**.

    Passcode is typically a long string (16-32 characters), which is provided
    in a form of QR image, but it can be obtained by requesting access to a string.

    Note that same code can be used to add a mobile phone as a duplicate authentication
    device at the same time when the same code is added into the Vault.

    **Robot framework example usage:**

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.MFA
        Library     RPA.Robocorp.Vault

        *** Tasks ***
        Generate time based code
            ${secrets}=    Get Secret   mfa
            ${code}=    Get Time Based OTP    ${secrets}[otpsecret]


    **Python example usage**

    .. code-block:: python

        from RPA.MFA import MFA
        from RPA.Robocorp.Vault import Vault


        def main():
            secrets = Vault().get_secret("mfa")
            code = MFA().get_time_based_otp(secrets["otpsecret"])
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(
        self,
        vault_name: Optional[str] = None,
        vault_key: Optional[str] = None,
        mode: Optional[OTPMode] = OTPMode.TIME,
    ):
        protect_keywords("RPA.MFA", ["get_oauth_token", "refresh_oauth_token"])
        self.logger = logging.getLogger(__name__)

        self._hotp: Optional[HOTP] = None
        self._totp: Optional[TOTP] = None
        if vault_name and vault_key:
            self.use_mfa_secret_from_vault(vault_name, vault_key, mode)
        self._oauth: Optional[OAuth2Session] = None

    @keyword
    def use_mfa_secret_from_vault(
        self, vault_name: str, vault_key: str, mode: OTPMode = OTPMode.TIME
    ):
        """Set `time` or `counter` based OTP with passcode stored in
        the Robocorp Vault named with `vault_name` under key of `vault_key`.

        :param vault_name: name of the vault storing the passcode
        :param vault_key: name of the vault key storing the passcode value
        """
        secrets = Vault().get_secret(vault_name)
        if mode == OTPMode.TIME:
            self.set_time_based_otp(secrets[vault_key])
        elif mode == OTPMode.COUNTER:
            self.set_counter_based_otp(secrets[vault_key])

    @keyword
    def set_time_based_otp(self, otp_passcode: str):
        """Set `time` based OTP with passcode.

        :param otp_passcode: the passcode provided by the Authenticator
        """
        self._totp = TOTP(otp_passcode)

    @keyword
    def set_counter_based_otp(self, otp_passcode: str):
        """Set `counter` based OTP with passcode.

        :param otp_passcode: the passcode provided by the Authenticator
        """
        self._hotp = HOTP(otp_passcode)

    @keyword
    def get_time_based_otp(self, otp_passcode: Optional[str] = None):
        """Get `time` based one time password using separately set
        passcode or by parameter `otp_passcode`.

        :param otp_passcode: the passcode provided by the Authenticator
        """
        if otp_passcode:
            self.set_time_based_otp(otp_passcode)
        if not self._totp:
            raise TOTPNotSetError(TOTPNotSetError.ERROR_MSG)
        return self._totp.now()

    @keyword
    def get_counter_based_otp(
        self,
        counter: int,
        otp_passcode: Optional[str] = None,
    ):
        """Get `counter` based one time password using separately set
        passcode or by parameter `otp_passcode`. The counter index is
        given by the `counter` parameter.

        :param counter: the index of the counter
        :param otp_passcode: the passcode provided by the Authenticator
        """
        if otp_passcode:
            self.set_counter_based_otp(otp_passcode)
        if not self._hotp:
            raise HOTPNotSetError(HOTPNotSetError.ERROR_MSG)
        return self._hotp.at(counter)

    @property
    def oauth(self) -> OAuth2Session:
        """Raises if there's no OAuth2 session already created."""
        if not self._oauth:
            raise OAuth2NotSetError(OAuth2NotSetError.ERROR_MSG)

        return self._oauth

    @keyword
    def generate_oauth_url(
        self, auth_url: str, *, client_id: str, redirect_uri: str, scope: str, **kwargs
    ) -> str:
        """Generates an authorization URL which must be opened by the user to start the
        OAuth2 flow and obtain an authorization code as response.

        The received response URL should be passed further with ``Get OAuth Token`` in
        order to complete the flow. Arbitrary keyword arguments can be passed to the
        keyword, which will be redirected to the wrapped `oauthlib` library method
        call.

        :param auth_url: Authorization endpoint to call the request on. (https URL
            usually ending with '/authorize')
        :param client_id: Client app ID. (generated by the provider)
        :param redirect_uri: Redirect URL allowed by the Client app configuration. (
            necessary for getting the `code` response)
        :param scope: Space-separated string of permissions. (accepted during the
            consent screen)
        :returns: Authorization URL string not containing any sensitive info in it.
            (call it with `access_type="offline"` or set the right `scope` in the
            authorization URL for ensuring the existence of the refresh token)

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Start OAuth Flow
                ${auth_url} =    Generate OAuth URL
                ...     https://accounts.google.com/o/oauth2/auth
                ...     client_id=810482312368-19htmcgcj*******googleusercontent.com
                ...     redirect_uri=https://developers.google.com/oauthplayground
                ...     scope=https://mail.google.com
                ...     access_type=offline     prompt=consent  # explicit grant
                Log     Start OAuth2 flow: ${auth_url}

        **Example: Python**

        .. code-block:: python

            from RPA.MFA import MFA

            lib_mfa = MFA()
            auth_url = lib_mfa.generate_oauth_url(
                "https://accounts.google.com/o/oauth2/auth", ...
            )
            print(f"Start OAuth2 flow: {auth_url}")
        """
        scopes: List[str] = scope.split()
        self._oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scopes)
        authorization_url, _ = self.oauth.authorization_url(auth_url, **kwargs)
        return authorization_url

    @keyword
    def get_oauth_token(
        self, token_url: str, *, client_secret: str, response_url: str, **kwargs
    ) -> dict:
        """Exchanges the code obtained previously with ``Generate OAuth URL`` for a
        token.

        The refresh token from the returned dictionary can be used further with the
        ``Refresh OAuth Token`` keyword in order to obtain a new access token when the
        previous one expires. (usually after one hour)
        Arbitrary keyword arguments can be passed to the keyword, which will be
        redirected to the wrapped `oauthlib` library method call.

        :param token_url: Token endpoint used with a POST request in order to retrieve
            the token data. (https URL usually ending with '/token')
        :param client_secret: Client app secret. (generated by the provider)
        :param response_url: The final URL containing the authorization `code` found in
            the address bar after authenticating and authorizing the Client app
            through the authorization URL.
        :returns: A dictionary containing the access token, metadata and optionally the
            refresh token.

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Finish OAuth Flow
                ${token} =      Get OAuth Token
                ...     https://accounts.google.com/o/oauth2/token
                ...     client_secret=GOCSPX-******mqZAW89
                ...     response_url=${resp_url}  # redirect of `Generate OAuth URL`

        **Example: Python**

        .. code-block:: python

            from RPA.MFA import MFA

            lib_mfa = MFA()
            lib_mfa.get_oauth_token("https://accounts.google.com/o/oauth2/token", ...)
        """
        token = self.oauth.fetch_token(
            token_url,
            client_secret=client_secret,
            authorization_response=response_url,
            **kwargs,
        )
        return dict(token)

    @keyword
    def refresh_oauth_token(
        self,
        token_url: str,
        *,
        client_id: Optional[str] = None,
        client_secret: str,
        refresh_token: Optional[str] = None,
        **kwargs,
    ) -> dict:
        """Refreshes the token as the access one usually expires after 1h and the
        refresh one never expires. (as long as it doesn't get revoked)

        The effect of this keyword is similar to ``Get OAuth Token``, but this time you
        refresh unattended an already existing token by receiving a new one instead.
        Arbitrary keyword arguments can be passed to the keyword, which will be
        redirected to the wrapped `oauthlib` library method call.

        :param token_url: Token endpoint used with a POST request in order to refresh
            the token data. (https URL usually ending with '/token')
        :param client_id: Client app ID. (generated by the provider)
        :param client_secret: Client app secret. (generated by the provider)
        :param refresh_token: Refresh token string found in the dictionary obtained
            with ``Get OAuth Token`` or ``Refresh OAuth Token``.
        :returns: A token dictionary containing a new access token and updated
            metadata. (the refresh token inside isn't guaranteed to remain constant)

        **Example: Robot Framework**

        .. code-block:: robotframework

            *** Tasks ***
            Refresh OAuth Flow
                ${token} =      Refresh OAuth Token
                ...     https://accounts.google.com/o/oauth2/token
                ...     client_id=810482312368-19htmcgcj*******googleusercontent.com
                ...     client_secret=GOCSPX-******mqZAW89
                ...     refresh_token=${token}[refresh_token]  # from `Get OAuth Token`

        **Example: Python**

        .. code-block:: python

            from RPA.MFA import MFA

            lib_mfa = MFA()
            lib_mfa.refresh_oauth_token(
                "https://accounts.google.com/o/oauth2/token", ...
            )
        """
        self._oauth = self._oauth or OAuth2Session(client_id)
        token = self.oauth.refresh_token(
            token_url,
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
            **kwargs,
        )
        return dict(token)



from pathlib import Path
import pytest
import subprocess
from time import sleep

from RPA.JavaAccessBridge import JavaAccessBridge


def open_test_application():
    TESTAPP_DIR = Path(__file__).resolve().parent / ".." / "resources" / "test-app"
    subprocess.Popen(
        ["java", "-jar", TESTAPP_DIR / "BasicSwing.jar"],
        shell=True,
        cwd=TESTAPP_DIR,
        close_fds=True,
    )


@pytest.mark.skip(reason="requires windows and java with UI")
def test_typing_text():
    jab = JavaAccessBridge()
    open_test_application()
    jab.select_window("Chat Frame")
    jab.type_text("role:text", "text for the textarea", enter=True)
    jab.type_text("role:text", "text for the input field", index=1, clear=True)
    jab.click_element("role:push button and name:Send")


if __name__ == "__main__":
    javalib = JavaAccessBridge()
    sleep(5)
    javalib.select_window("Chat Frame")
    sleep(5)
    javalib.close_java_window()



from json.encoder import JSONEncoder
import time
from typing import List
from urllib.parse import parse_qs, urlencode, urlparse
from mock import MagicMock, ANY
import pytest
from pytest_mock import MockerFixture
from RPA.MSGraph import MSGraph, DEFAULT_REDIRECT_URI
from O365.sharepoint import Site
from pathlib import Path
import re

RESOURCE_DIR = Path(__file__).parent / "resources"
TEMP_DIR = Path(__file__).parent.parent / "results"
CONFIG_FILE = RESOURCE_DIR / "msgraph"
SCOPES = MSGraph().get_scopes()

DEFAULT_STATE = "123"
MOCK_CLIENT_ID = "my-client-id"
MOCK_CLIENT_SECRET = "my-client-secret"
MOCK_AUTH_CODE = "https://localhost/myapp/?code=my-mock-auth-code-123&state={}&session_state=mock-session-state#"
MOCK_ACCESS_TOKEN = "microsoft-access-token-{:0>2}"
MOCK_REFRESH_TOKEN = "microsoft-refresh-token-{:0>2}"


@pytest.fixture
def library() -> MSGraph:
    return MSGraph(file_backend_path=TEMP_DIR)


@pytest.fixture
def configured_lib(library: MSGraph) -> MSGraph:
    library.configure_msgraph_client(MOCK_CLIENT_ID, MOCK_CLIENT_SECRET)
    return library


@pytest.fixture
def init_auth(library: MSGraph, mocker: MockerFixture) -> str:
    return library.generate_oauth_authorization_url(MOCK_CLIENT_ID, MOCK_CLIENT_SECRET)


def _get_stateful_mock_auth_code(init_auth: str) -> str:
    init_query = parse_qs(urlparse(init_auth).query)
    return MOCK_AUTH_CODE.format(init_query["state"][0])


@pytest.fixture
def authorized_lib(
    configured_lib: MSGraph,
    mocker: MockerFixture,
    init_auth: str,
) -> MSGraph:
    _patch_token_response(configured_lib, mocker, 1)
    configured_lib.authorize_and_get_token(_get_stateful_mock_auth_code(init_auth))
    return configured_lib


@pytest.fixture
def sharepoint_site(authorized_lib: MSGraph, mocker: MockerFixture) -> Site:
    site_id = "contoso.sharepoint.com"
    response = {
        "id": "contoso.sharepoint.com,2C712604-1370-44E7-A1F5-426573FDA80A,2D2244C3-251A-49EA-93A8-39E1C3A060FE",
        "displayName": "OneDrive Team Site",
        "name": "1drvteam",
        "createdDateTime": "2017-05-09T20:56:00Z",
        "lastModifiedDateTime": "2017-05-09T20:56:01Z",
        "webUrl": "https://contoso.sharepoint.com/teams/1drvteam",
    }
    _patch_graph_response(authorized_lib, mocker, response)

    return authorized_lib.get_sharepoint_site(site_id)


def _patch_token_response(
    library: MSGraph, mocker: MockerFixture, iteration: int
) -> MockerFixture._Patcher:
    return _patch_graph_response(
        library,
        mocker,
        {
            "token_type": "Bearer",
            "scope": "%20F".join(SCOPES),
            "expires_in": 3600,
            "access_token": MOCK_ACCESS_TOKEN.format(iteration),
            "refresh_token": MOCK_REFRESH_TOKEN.format(iteration),
        },
    )


def _create_graph_json_response(return_value: dict) -> MagicMock:
    mock_graph_response = MagicMock()
    mock_graph_response.status_code = 200
    mock_graph_response.headers = {"Content-Type": "application/json"}
    mock_graph_response.json.return_value = return_value
    mock_graph_response.text = JSONEncoder().encode(return_value)
    return mock_graph_response


def _patch_graph_response(
    library: MSGraph, mocker: MockerFixture, return_value: dict
) -> MagicMock:
    mock_graph_response = _create_graph_json_response(return_value)
    config = {"return_value": mock_graph_response}

    return mocker.patch.object(library.client.connection.session, "request", **config)


def _patch_multiple_graph_responses(
    library: MSGraph, mocker: MockerFixture, mocked_responses: List[MagicMock]
) -> MagicMock:
    config = {"side_effect": mocked_responses}

    return mocker.patch.object(library.client.connection.session, "request", **config)


def test_configuring_graph_client(library: MSGraph, mocker: MockerFixture) -> None:
    mock_client = mocker.patch("RPA.MSGraph.Account", autospec=True)

    library.generate_oauth_authorization_url(MOCK_CLIENT_ID, MOCK_CLIENT_SECRET)

    mock_client.assert_any_call((MOCK_CLIENT_ID, MOCK_CLIENT_SECRET), token_backend=ANY)


def test_generating_auth_url(init_auth: str) -> None:
    params = {
        "response_type": "code",
        "client_id": MOCK_CLIENT_ID,
        "redirect_uri": DEFAULT_REDIRECT_URI,
        "scope": " ".join(SCOPES),
    }
    encoded_params = urlencode(params).replace(r"+", r"\+")
    pattern = re.compile(
        rf"https:\/\/login.microsoftonline.com\/common\/oauth2\/v2.0\/authorize\?{encoded_params}"
        r"&state=[a-zA-Z0-9]*&access_type=offline"
    )
    result = re.match(pattern, init_auth)
    assert result


def test_auth_cycle(
    library: MSGraph,
    mocker: MockerFixture,
    init_auth: str,
) -> None:
    _patch_token_response(library, mocker, 1)

    refresh_token = library.authorize_and_get_token(
        _get_stateful_mock_auth_code(init_auth)
    )

    assert library.token_backend.get_token()[
        "access_token"
    ] == MOCK_ACCESS_TOKEN.format(1)
    assert refresh_token == MOCK_REFRESH_TOKEN.format(1)


def test_refreshing_token(configured_lib: MSGraph, mocker: MockerFixture) -> None:
    return_token = {
        "token_type": "Bearer",
        "expires_in": 3600,
        "access_token": MOCK_ACCESS_TOKEN.format(2),
        "refresh_token": MOCK_REFRESH_TOKEN.format(2),
        "expires_at": time.time() + 3600,
        "scope": "%20F".join(SCOPES),
        "scopes": SCOPES,
    }

    config = {"return_value.refresh_token.return_value": return_token}
    mocker.patch("O365.connection.OAuth2Session", **config)

    refresh_token = configured_lib.refresh_oauth_token(MOCK_REFRESH_TOKEN.format(1))

    assert configured_lib.token_backend.get_token()[
        "access_token"
    ] == MOCK_ACCESS_TOKEN.format(2)
    assert refresh_token == MOCK_REFRESH_TOKEN.format(2)


def test_get_me(authorized_lib: MSGraph, mocker: MockerFixture) -> None:
    data = {
        "businessPhones": ["+1 425 555 0109"],
        "displayName": "Adele Vance",
        "givenName": "Adele",
        "jobTitle": "Retail Manager",
        "mail": "AdeleV@contoso.onmicrosoft.com",
        "mobilePhone": "+1 425 555 0109",
        "officeLocation": "18/2111",
        "preferredLanguage": "en-US",
        "surname": "Vance",
        "userPrincipalName": "AdeleV@contoso.onmicrosoft.com",
        "id": "87d349ed-44d7-43e1-9a83-5f2406dee5bd",
    }
    m = _patch_graph_response(authorized_lib, mocker, data)

    user_me = authorized_lib.get_me()

    m.assert_called_once()
    assert str(user_me) == data["displayName"]
    assert user_me.object_id == data["id"]


@pytest.mark.parametrize(
    "search_string,response",
    [
        (
            "adam",
            {
                "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users",
                "value": [
                    {
                        "businessPhones": [],
                        "displayName": "Conf Room Adams",
                        "givenName": None,
                        "jobTitle": None,
                        "mail": "Adams@contoso.com",
                        "mobilePhone": None,
                        "officeLocation": None,
                        "preferredLanguage": None,
                        "surname": None,
                        "userPrincipalName": "Adams@contoso.com",
                        "id": "6ea91a8d-e32e-41a1-b7bd-d2d185eed0e0",
                    },
                    {
                        "businessPhones": ["425-555-0100"],
                        "displayName": "Adam Administrator",
                        "givenName": "Adam-adm",
                        "jobTitle": None,
                        "mail": None,
                        "mobilePhone": "425-555-0101",
                        "officeLocation": None,
                        "preferredLanguage": "en-US",
                        "surname": "Administrator",
                        "userPrincipalName": "admin@contoso.com",
                        "id": "4562bcc8-c436-4f95-b7c0-4f8ce89dca5e",
                    },
                ],
            },
        ),
        (
            "john",
            {
                "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users",
                "value": [
                    {
                        "businessPhones": ["555-555-0100"],
                        "displayName": "Johnny Apple",
                        "givenName": "John",
                        "jobTitle": "IT Admin",
                        "mail": "j.apple@contoso.com",
                        "mobilePhone": None,
                        "officeLocation": None,
                        "preferredLanguage": None,
                        "surname": None,
                        "userPrincipalName": "j.apple@contoso.com",
                        "id": "6ea91a8d-e32e-41a1-b7bd-d2d185eed123",
                    },
                    {
                        "businessPhones": ["555-123-0100"],
                        "displayName": "John Smith",
                        "givenName": "John",
                        "jobTitle": "BDR",
                        "mail": "j.smith@contoso.com",
                        "mobilePhone": "555-123-0101",
                        "officeLocation": None,
                        "preferredLanguage": "en-US",
                        "surname": "Administrator",
                        "userPrincipalName": "admin@contoso.com",
                        "id": "4562bcc8-c436-4f95-b7c0-4f8ce89dc123",
                    },
                ],
            },
        ),
    ],
)
def test_search_for_users(
    authorized_lib: MSGraph, mocker: MockerFixture, search_string: str, response: dict
) -> None:
    m = _patch_graph_response(authorized_lib, mocker, response)

    users = authorized_lib.search_for_users(search_string)

    m.assert_called_once()
    for user in users:
        assert user.display_name in [u["displayName"] for u in response["value"]]
        assert user.user_principal_name in [
            u["userPrincipalName"] for u in response["value"]
        ]


@pytest.mark.parametrize(
    "folder_path,responses",
    [
        (
            "/Path/To/Folder/",
            [
                {
                    "createdBy": {
                        "user": {
                            "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                            "displayName": "Ryan Gregg",
                        }
                    },
                    "createdDateTime": "2016-03-21T20:01:37Z",
                    "cTag": '"c:{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},0"',
                    "eTag": '"{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},1"',
                    "folder": {"childCount": 120},
                    "id": "01NKDM7HMOJTVYMDOSXFDK2QJDXCDI3WUK",
                    "lastModifiedBy": {
                        "user": {
                            "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                            "displayName": "Ryan Gregg",
                        }
                    },
                    "lastModifiedDateTime": "2016-03-21T20:01:37Z",
                    "name": "OneDrive",
                    "root": {},
                    "size": 157286400,
                    "webUrl": "https://contoso-my.sharepoint.com/personal/rgregg_contoso_com/Documents",
                },
                {
                    "value": [
                        {"name": "myfile.jpg", "size": 2048, "file": {}},
                        {"name": "Documents", "folder": {"childCount": 4}},
                        {"name": "Photos", "folder": {"childCount": 203}},
                        {"name": "my sheet(1).xlsx", "size": 197},
                    ],
                },
            ],
        )
    ],
)
def test_listing_files_onedrive_folder(
    authorized_lib: MSGraph,
    mocker: MockerFixture,
    folder_path: str,
    responses: List[dict],
) -> None:
    mocked_responses = [_create_graph_json_response(r) for r in responses]
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    items = authorized_lib.list_files_in_onedrive_folder(folder_path)

    files_in_response = [
        item for item in responses[1]["value"] if not item.get("folder")
    ]
    for i, item in enumerate(items):
        assert item.name == files_in_response[i]["name"]
        assert not item.is_folder


@pytest.mark.parametrize(
    "file_path,responses",
    [
        (
            "/Path/To/File",
            [
                {
                    "name": "my notes.txt",
                    "size": 197,
                    "id": "01NKDM7HMOJTVYMDOSXFDK2QJDXCDI3WUK",
                },
                "dummy file content".encode(),
            ],
        )
    ],
)
def test_downloading_file_from_onedrive(
    authorized_lib: MSGraph,
    mocker: MockerFixture,
    file_path: str,
    responses: List[dict],
) -> None:
    mocked_responses = []
    mocked_responses.append(_create_graph_json_response(responses[0]))
    mocked_response = MagicMock()
    mocked_response.__enter__.return_value.status_code = 200
    mocked_response.__enter__.return_value.headers = {
        "Content-Type": "application/octet-stream"
    }
    mocked_response.__enter__.return_value.content = responses[1]
    mocked_responses.append(mocked_response)
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    downloaded_file = authorized_lib.download_file_from_onedrive(file_path, TEMP_DIR)

    assert downloaded_file
    assert downloaded_file.exists()


@pytest.mark.parametrize(
    "file_path,responses",
    [
        (
            "/Path/To/File",
            [
                {
                    "createdBy": {
                        "user": {
                            "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                            "displayName": "Ryan Gregg",
                        }
                    },
                    "createdDateTime": "2016-03-21T20:01:37Z",
                    "cTag": '"c:{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},0"',
                    "eTag": '"{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},1"',
                    "folder": {"childCount": 120},
                    "id": "01NKDM7HMOJTVYMDOSXFDK2QJDXCDI3WUK",
                    "lastModifiedBy": {
                        "user": {
                            "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                            "displayName": "Ryan Gregg",
                        }
                    },
                    "lastModifiedDateTime": "2016-03-21T20:01:37Z",
                    "name": "OneDrive",
                    "root": {},
                    "size": 157286400,
                    "webUrl": "https://contoso-my.sharepoint.com/personal/rgregg_contoso_com/Documents",
                },
                {
                    "value": [
                        {"name": "myfile.jpg", "size": 2048, "file": {}},
                        {"name": "my sheet(1).xlsx", "size": 197},
                    ],
                },
                "dummy file content".encode(),
            ],
        )
    ],
)
def test_downloading_folder_from_onedrive(
    authorized_lib: MSGraph,
    mocker: MockerFixture,
    file_path: str,
    responses: List[dict],
) -> None:
    mocked_responses = [
        _create_graph_json_response(r) for r in (responses[0], responses[1])
    ]

    mocked_response = MagicMock()
    mocked_response.__enter__.return_value.status_code = 200
    mocked_response.__enter__.return_value.headers = {
        "Content-Type": "application/octet-stream"
    }
    mocked_response.__enter__.return_value.content = responses[2]
    for _ in responses[1]["value"]:
        mocked_responses.append(mocked_response)

    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    downloaded_folder = authorized_lib.download_folder_from_onedrive(
        file_path, TEMP_DIR / "downloaded_folder"
    )

    assert downloaded_folder
    assert downloaded_folder.exists()


@pytest.mark.parametrize(
    "search_string,response",
    [
        (
            "Contoso Project",
            {
                "value": [
                    {
                        "id": "0123456789abc!123",
                        "name": "Contoso Project",
                        "folder": {},
                        "searchResult": {
                            "onClickTelemetryUrl": "https://bing.com/0123456789abc!123"
                        },
                    },
                    {
                        "id": "0123456789abc!456",
                        "name": "Contoso Project 2016",
                        "folder": {},
                        "searchResult": {
                            "onClickTelemetryUrl": "https://bing.com/0123456789abc!456"
                        },
                    },
                ],
            },
        )
    ],
)
def test_finding_onedrive_file(
    authorized_lib: MSGraph, mocker: MockerFixture, search_string: str, response: dict
) -> None:
    m = _patch_graph_response(authorized_lib, mocker, response)

    items = authorized_lib.find_onedrive_file(search_string)

    m.assert_called_once()
    for item in items:
        assert item.object_id in [i["id"] for i in response["value"]]
        assert item.name in [i["name"] for i in response["value"]]


@pytest.mark.parametrize(
    "share_url,responses",
    [
        (
            "https://1drv.ms/v/s!AjonToUPWqXmgjO3RqDbhRaSMrOM",
            [
                {
                    "name": "report.txt",
                    "size": 1997,
                    "id": "01NKDM7HMOJTVYMDOSXFDK2QJDXCDI3WUK",
                },
                "dummy file content".encode(),
            ],
        )
    ],
)
def test_downloading_from_share_link(
    authorized_lib: MSGraph,
    mocker: MockerFixture,
    share_url: str,
    responses: List[dict],
) -> None:
    mocked_responses = []
    mocked_responses.append(_create_graph_json_response(responses[0]))
    mocked_response = MagicMock()
    mocked_response.__enter__.return_value.status_code = 200
    mocked_response.__enter__.return_value.headers = {
        "Content-Type": "application/octet-stream"
    }
    mocked_response.__enter__.return_value.content = responses[1]
    mocked_responses.append(mocked_response)
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    downloaded_file = authorized_lib.download_file_from_share_link(share_url, TEMP_DIR)

    assert downloaded_file
    assert downloaded_file.exists()


@pytest.mark.parametrize(
    "file_path,folder_path,responses",
    [
        (
            "/Path/To/my notes.txt",
            "/Path/To/Folder/",
            [
                {
                    "createdBy": {
                        "user": {
                            "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                            "displayName": "Ryan Gregg",
                        }
                    },
                    "createdDateTime": "2016-03-21T20:01:37Z",
                    "cTag": '"c:{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},0"',
                    "eTag": '"{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},1"',
                    "folder": {"childCount": 120},
                    "id": "01NKDM7HMOJTVYMDOSXFDK2QJDXCDI3WUK",
                    "lastModifiedBy": {
                        "user": {
                            "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                            "displayName": "Ryan Gregg",
                        }
                    },
                    "lastModifiedDateTime": "2016-03-21T20:01:37Z",
                    "name": "OneDrive",
                    "root": {},
                    "size": 157286400,
                    "webUrl": "https://contoso-my.sharepoint.com/personal/rgregg_contoso_com/Documents",
                },
                {
                    "name": "my notes.txt",
                    "size": 197,
                    "id": "01NKDM7HMOJTVYMDOSXFDK2QJDXCDI3WUK",
                },
            ],
        )
    ],
)
def test_uploading_file_to_onedrive(
    authorized_lib: MSGraph,
    mocker: MockerFixture,
    file_path: str,
    folder_path: str,
    responses: List[dict],
) -> None:
    mocked_responses = [_create_graph_json_response(r) for r in responses]
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    # Mock file interactions
    mocker.patch("pathlib.Path.exists", return_value=True)
    mocker.patch("pathlib.Path.is_file", return_value=True)
    mocked_path = MagicMock()
    mocked_path.st_size = 956514
    mocker.patch("pathlib.Path.stat", return_value=mocked_path)
    mocker.patch("io.open", mocker.mock_open(read_data="secret notes"))

    item = authorized_lib.upload_file_to_onedrive(file_path, folder_path)

    assert item.name == responses[1]["name"]


@pytest.mark.parametrize(
    "args",
    [
        ("root"),
        ("contoso.sharepoint.com"),
        ("contoso.sharepoint.com", "/path/to/site"),
        (
            "contoso.sharepoint.com,a384ebb0-67a5-4976-a7eb-60bbd5d5f87a,f0a2c07f-bfa0-4259-b9cd-fbdab88ebcf4"
        ),
        (
            "contoso.sharepoint.com",
            "a384ebb0-67a5-4976-a7eb-60bbd5d5f87a",
            "f0a2c07f-bfa0-4259-b9cd-fbdab88ebcf4",
        ),
    ],
)
def test_get_sharepoint_site(
    authorized_lib: MSGraph, mocker: MockerFixture, args: tuple
) -> None:
    response = {
        "id": "contoso.sharepoint.com,2C712604-1370-44E7-A1F5-426573FDA80A,2D2244C3-251A-49EA-93A8-39E1C3A060FE",
        "displayName": "OneDrive Team Site",
        "name": "1drvteam",
        "createdDateTime": "2017-05-09T20:56:00Z",
        "lastModifiedDateTime": "2017-05-09T20:56:01Z",
        "webUrl": "https://contoso.sharepoint.com/teams/1drvteam",
    }
    m = _patch_graph_response(authorized_lib, mocker, response)

    if isinstance(args, str):
        site = authorized_lib.get_sharepoint_site(args)
    else:
        site = authorized_lib.get_sharepoint_site(*args)

    m.assert_called_once()
    assert site.display_name == response["displayName"]
    assert site.object_id == response["id"]


def test_get_items_from_sharepoint_list(
    authorized_lib: MSGraph, mocker: MockerFixture, sharepoint_site: Site
) -> None:
    list_name = "Documents"
    list_response = {
        "id": "b57af081-936c-4803-a120-d94887b03864",
        "name": "Documents",
        "createdDateTime": "2016-08-30T08:32:00Z",
        "lastModifiedDateTime": "2016-08-30T08:32:00Z",
        "list": {"hidden": False, "template": "documentLibrary"},
    }
    columns_response = {
        "value": [
            {
                "description": "",
                "displayName": "Name",
                "hidden": False,
                "id": "99ddcf45-e2f7-4f17-82b0-6fba34445103",
                "indexed": False,
                "name": "Name",
                "readOnly": False,
                "required": False,
                "text": {
                    "allowMultipleLines": False,
                    "appendChangesToExistingText": False,
                    "linesForEditing": 0,
                    "maxLength": 255,
                },
            },
            {
                "description": "",
                "displayName": "Color",
                "id": "11dfef35-e2f7-4f17-82b0-6fba34445103",
                "indexed": False,
                "name": "Color",
                "readOnly": False,
                "required": False,
                "text": {
                    "allowMultipleLines": False,
                    "appendChangesToExistingText": False,
                    "linesForEditing": 0,
                    "maxLength": 255,
                },
            },
            {
                "description": "",
                "displayName": "Quantity",
                "id": "27c36545-4c19-4af8-a5a1-eaf520dbba25",
                "indexed": False,
                "name": "Quantity",
                "readOnly": False,
                "required": False,
                "text": {
                    "allowMultipleLines": False,
                    "appendChangesToExistingText": False,
                    "linesForEditing": 0,
                    "maxLength": 255,
                },
            },
        ]
    }
    list_items_response = {
        "value": [
            {
                "id": "2",
                "fields": {"Name": "Gadget", "Color": "Red", "Quantity": 503},
            },
            {
                "id": "4",
                "fields": {"Name": "Widget", "Color": "Blue", "Quantity": 2357},
            },
            {
                "id": "7",
                "fields": {"Name": "Gizmo", "Color": "Green", "Quantity": 92},
            },
        ]
    }
    mocked_responses = [
        _create_graph_json_response(r)
        for r in (list_response, columns_response, list_items_response)
    ]
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    items = authorized_lib.get_items_from_sharepoint_list(list_name, sharepoint_site)

    assert items
    for item in items:
        assert item["object_id"] in [i["id"] for i in list_items_response["value"]]


def test_create_sharepoint_list(
    authorized_lib: MSGraph, mocker: MockerFixture, sharepoint_site: Site
) -> None:
    new_list_data = {
        "displayName": "Books",
        "columns": [
            {"name": "Author", "text": {}},
            {"name": "PageCount", "number": {}},
        ],
        "list": {"template": "genericList"},
    }
    list_response = {
        "id": "22e03ef3-6ef4-424d-a1d3-92a337807c30",
        "createdDateTime": "2017-04-30T01:21:00Z",
        "createdBy": {
            "user": {
                "displayName": "Ryan Gregg",
                "id": "8606e4d5-d582-4f5f-aeba-7d7c18b20cfd",
            }
        },
        "lastModifiedDateTime": "2016-08-30T08:26:00Z",
        "lastModifiedBy": {
            "user": {
                "displayName": "Ryan Gregg",
                "id": "8606e4d5-d582-4f5f-aeba-7d7c18b20cfd",
            }
        },
    }
    columns_response = {
        "value": [
            {
                "description": "",
                "displayName": "Author",
                "hidden": False,
                "id": "99ddcf45-e2f7-4f17-82b0-6fba34445103",
                "indexed": False,
                "name": "Author",
                "readOnly": False,
                "required": False,
                "text": {
                    "allowMultipleLines": False,
                    "appendChangesToExistingText": False,
                    "linesForEditing": 0,
                    "maxLength": 255,
                },
            },
            {
                "description": "",
                "displayName": "PageCount",
                "id": "11dfef35-e2f7-4f17-82b0-6fba34445103",
                "indexed": False,
                "name": "PageCount",
                "readOnly": False,
                "required": False,
                "text": {
                    "allowMultipleLines": False,
                    "appendChangesToExistingText": False,
                    "linesForEditing": 0,
                    "maxLength": 255,
                },
            },
        ]
    }
    mocked_responses = [
        _create_graph_json_response(r) for r in (list_response, columns_response)
    ]
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    sp_list = authorized_lib.create_sharepoint_list(new_list_data, sharepoint_site)

    assert sp_list.object_id == list_response["id"]
    assert (
        sp_list.created_by.display_name
        == list_response["createdBy"]["user"]["displayName"]
    )


def test_list_sharepoint_drives(
    authorized_lib: MSGraph, mocker: MockerFixture, sharepoint_site: Site
) -> None:
    response = {
        "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#drives",
        "value": [
            {
                "createdDateTime": "2017-07-27T02:41:36Z",
                "description": "",
                "id": "b!-RIj2DuyvEyV1T4NlOaMHk8XkS_I8MdFlUCq1BlcjgmhRfAj3-Z8RY2VpuvV_tpd",
                "lastModifiedDateTime": "2018-03-27T07:34:38Z",
                "name": "OneDrive",
                "webUrl": "https://m365x214355-my.sharepoint.com/personal/meganb_m365x214355_onmicrosoft_com/Documents",
                "driveType": "business",
                "createdBy": {"user": {"displayName": "System Account"}},
                "lastModifiedBy": {
                    "user": {
                        "email": "MeganB@M365x214355.onmicrosoft.com",
                        "id": "48d31887-5fad-4d73-a9f5-3c356e68a038",
                        "displayName": "Megan Bowen",
                    }
                },
                "owner": {
                    "user": {
                        "email": "MeganB@M365x214355.onmicrosoft.com",
                        "id": "48d31887-5fad-4d73-a9f5-3c356e68a038",
                        "displayName": "Megan Bowen",
                    }
                },
                "quota": {
                    "deleted": 0,
                    "remaining": 1099217021300,
                    "state": "normal",
                    "total": 1099511627776,
                    "used": 294606476,
                },
            }
        ],
    }
    m = _patch_graph_response(authorized_lib, mocker, response)

    sp_drives = authorized_lib.list_sharepoint_site_drives(sharepoint_site)

    m.assert_called_once()
    assert sp_drives
    for sp_drive in sp_drives:
        assert sp_drive.object_id in [drive["id"] for drive in response["value"]]
        assert sp_drive.name in [drive["name"] for drive in response["value"]]


def test_list_files_in_sharepoint_drive(
    authorized_lib: MSGraph,
    mocker: MockerFixture,
    sharepoint_site: Site,
) -> None:
    drive_id = None
    response_folder = {
        "createdBy": {
            "user": {
                "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                "displayName": "Ryan Gregg",
            }
        },
        "createdDateTime": "2016-03-21T20:01:37Z",
        "cTag": '"c:{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},0"',
        "eTag": '"{86EB4C8E-D20D-46B9-AD41-23B8868DDA8A},1"',
        "folder": {"childCount": 120},
        "id": "01NKDM7HMOJTVYMDOSXFDK2QJDXCDI3WUK",
        "lastModifiedBy": {
            "user": {
                "id": "efee1b77-fb3b-4f65-99d6-274c11914d12",
                "displayName": "Ryan Gregg",
            }
        },
        "lastModifiedDateTime": "2016-03-21T20:01:37Z",
        "name": "OneDrive",
        "root": {},
        "size": 157286400,
        "webUrl": "https://contoso-my.sharepoint.com/personal/rgregg_contoso_com/Documents",
    }

    response_files = {
        "value": [
            {"name": "myfile.jpg", "size": 2048, "file": {}},
            {"name": "Documents", "folder": {"childCount": 4}},
            {"name": "Photos", "folder": {"childCount": 203}},
            {"name": "my sheet(1).xlsx", "size": 197},
        ],
    }
    mocked_responses = [
        _create_graph_json_response(r) for r in (response_folder, response_files)
    ]
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    sp_files = authorized_lib.list_files_in_sharepoint_site_drive(
        sharepoint_site, drive=drive_id
    )

    for file in sp_files:
        assert file.name in [file["name"] for file in response_files["value"]]


@pytest.mark.parametrize(
    "file_path, target_directory, drive_id, responses",
    [
        (
            "/Path/To/File",
            TEMP_DIR,
            None,
            [
                {
                    "name": "file-from-sharepoint.txt",
                    "size": 500097,
                    "id": "3JL14H5L1HJC0PPPQO7EMQB",
                },
                "dummy file content".encode(),
            ],
        ),
    ],
)
def test_download_file_from_sharepoint(
    authorized_lib: MSGraph,
    mocker: MockerFixture,
    sharepoint_site: Site,
    drive_id: str,
    file_path: str,
    target_directory: str,
    responses: List[dict],
) -> None:
    mocked_responses = []
    mocked_responses.append(_create_graph_json_response(responses[0]))
    mocked_response = MagicMock()
    mocked_response.__enter__.return_value.status_code = 200
    mocked_response.__enter__.return_value.headers = {
        "Content-Type": "application/octet-stream"
    }
    mocked_response.__enter__.return_value.content = responses[1]
    mocked_responses.append(mocked_response)
    _patch_multiple_graph_responses(authorized_lib, mocker, mocked_responses)

    downloaded_file = authorized_lib.download_file_from_sharepoint(
        file_path, sharepoint_site, target_directory, drive=drive_id
    )

    assert downloaded_file
    assert downloaded_file.exists()



import atexit
import logging
import platform
import struct
import time
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Any, List

from RPA.Email.common import counter_duplicate_path

if platform.system() == "Windows":
    import win32api
    import win32com.client
    import pywintypes
else:
    logging.getLogger(__name__).warning(
        "RPA.Outlook.Application library works only on Windows platform"
    )


def _to_unsigned(val):
    return struct.unpack("L", struct.pack("l", val))[0]


@contextmanager
def catch_com_error():
    """Try to convert COM errors to human readable format."""
    try:
        yield
    except pywintypes.com_error as err:  # pylint: disable=no-member
        if err.excepinfo:
            try:
                msg = win32api.FormatMessage(_to_unsigned(err.excepinfo[5]))
            except Exception:  # pylint: disable=broad-except
                msg = err.excepinfo[2]
        else:
            try:
                msg = win32api.FormatMessage(_to_unsigned(err.hresult))
            except Exception:  # pylint: disable=broad-except
                msg = err.strerror
        raise RuntimeError(msg) from err


class Application:
    # pylint: disable=C0301
    """`Outlook.Application` is a library for controlling the Outlook application.

    *Note*. Library works only Windows platform.

    **About Email Filtering**

    Emails can be filtered according to specification set by Restrict method of the Item
    class https://docs.microsoft.com/en-us/office/vba/api/outlook.items.restrict.

    Couple of examples:

    .. code-block:: robotframework

        Get Emails
        ...   email_filter=[Subject]='test email'

        Move Emails
        ...   email_filter=[SenderEmailAddress]='hello@gmail.com'

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library                 RPA.Outlook.Application
        Task Setup              Open Application
        Suite Teardown          Quit Application

        *** Variables ***
        ${RECIPIENT}            address@domain.com

        *** Tasks ***
        Send message
            Send Message       recipients=${RECIPIENT}
            ...                subject=This is the subject
            ...                body=This is the message body
            ..                 attachments=approved.png

    **Python**

    .. code-block:: python

        from RPA.Outlook.Application import Application

        def send_message():
            app = Application()
            app.open_application()
            app.send_message(
                recipients='EMAILADDRESS_1, EMAILADDRESS_2',
                subject='email subject',
                body='email body message',
                attachments='../orders.csv'

    For more information, see: https://docs.microsoft.com/en-us/previous-versions/office/developer/office-2007/bb219950(v=office.12)
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self, autoexit: bool = True) -> None:
        self.logger = logging.getLogger(__name__)
        self.app = None

        if platform.system() != "Windows":
            self.logger.warning(
                "Outlook application library requires Windows dependencies to work."
            )
        if autoexit:
            atexit.register(self.quit_application)

    def open_application(
        self, visible: bool = False, display_alerts: bool = False
    ) -> None:
        """Open the Outlook application.

        :param visible: show window after opening, default False
        :param display_alerts: show alert popups, default False
        """
        with catch_com_error():
            self.app = win32com.client.gencache.EnsureDispatch("Outlook.Application")

            if hasattr(self.app, "Visible"):
                self.app.Visible = visible

            # show eg. file overwrite warning or not
            if hasattr(self.app, "DisplayAlerts"):
                self.app.DisplayAlerts = display_alerts

    def close_document(self, save_changes: bool = False) -> None:
        """Close the active document (if open).

        :param save_changes: if changes should be saved on close, default False
        """
        if not self.app:
            return
        if hasattr(self.app, "ActiveDocument"):
            self.app.ActiveDocument.Close(save_changes)

    def quit_application(self, save_changes: bool = False) -> None:
        """Quit the application.

        :param save_changes: if changes should be saved on quit, default False
        """
        if not self.app:
            return
        self.close_document(save_changes)
        self.app.Quit()
        self.app = None

    def send_message(
        self,
        recipients: Any,
        subject: str,
        body: str,
        html_body: bool = False,
        attachments: Any = None,
        save_as_draft: bool = False,
    ) -> bool:
        """Send message with Outlook

        :param recipients: list of addresses
        :param subject: email subject
        :param body: email body
        :param html_body: True if body contains HTML, defaults to False
        :param attachments: list of filepaths to include in the email, defaults to []
        :param save_as_draft: message is saved as draft when `True`
         instead (and not sent)
        :return: `True` if there were no errors
        """
        self.logger.warning(
            "Keyword 'Send Message' is deprecated, "
            "and will be removed in a future version."
            "Use 'Send Email' instead."
        )
        return self.send_email(
            recipients, subject, body, html_body, attachments, save_as_draft
        )

    def send_email(
        self,
        recipients: Any,
        subject: str,
        body: str,
        html_body: bool = False,
        attachments: Any = None,
        save_as_draft: bool = False,
    ) -> bool:
        """Send email with Outlook

        :param recipients: list of addresses
        :param subject: email subject
        :param body: email body
        :param html_body: True if body contains HTML, defaults to False
        :param attachments: list of filepaths to include in the email, defaults to []
        :param save_as_draft: email is saved as draft when `True`
        :return: `True` if there were no errors

        Example:

        .. code-block:: robotframework

            Send Email
            ...    recipients=robocorp.tester@gmail.com
            ...    subject=hello from Outlook
            ...    body=empty body
            ...    attachments=${CURDIR}${/}example.xlsx
        """
        # pylint: disable=no-member
        attachments = attachments or []
        if not isinstance(recipients, list):
            recipients = recipients.split(",")
        if not isinstance(attachments, list):
            attachments = str(attachments).split(",")

        mailto = ";".join(recipients)

        try:
            mail = self.app.CreateItem(0)
            mail.To = mailto
            mail.Subject = subject

            if html_body:
                mail.HTMLBody = body
            else:
                mail.Body = body

            self._add_attachments(mail, attachments)

            if save_as_draft:
                mail.Save()
                self.logger.debug("Email draft saved")
            else:
                mail.Send()
                self.logger.debug("Email sent")
        except pywintypes.com_error as e:
            self.logger.error(
                f"Mail {'saving' if save_as_draft else 'sending'} failed: %s", e
            )
            return False
        return True

    def _add_attachments(self, email, attachments):
        for attachment in attachments:
            filepath = Path(attachment).absolute()
            email.Attachments.Add(str(filepath))

    def _is_email_too_old(self, email):
        now = datetime.now().timestamp()
        msg_time = email["ReceivedTimestamp"]
        timediff = now - msg_time
        return timediff > 30.0

    def _check_for_matching(self, criterion, email):
        crit_key, crit_val = criterion.split(":", 1)
        crit_key = crit_key.upper()
        crit_val = crit_val.lower()
        match_found = None

        if crit_key.upper() == "SUBJECT" and crit_val in email["Subject"].lower():
            match_found = email
        elif crit_key.upper() == "SENDER" and crit_val in email["Sender"].lower():
            match_found = email
        elif crit_key.upper() == "BODY" and crit_val in email["Body"].lower():
            match_found = email

        return match_found

    def wait_for_message(
        self, criterion: str = None, timeout: float = 5.0, interval: float = 1.0
    ) -> Any:
        """Wait for email matching `criterion` to arrive into mailbox.

        :param criterion: message filter to wait for, defaults to ""
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0
        :return: list of messages or False

        Possible wait criterias are: SUBJECT, SENDER and BODY

        Example:

        .. code-block:: robotframework

            Wait for message     SUBJECT:rpa task calling    timeout=300    interval=10
        """
        self.logger.warning(
            "Keyword 'Wait For Message' is deprecated, "
            "and will be removed in a future version."
            "Use 'Wait For Email' instead."
        )
        return self.wait_for_email(criterion, timeout, interval)

    def wait_for_email(
        self, criterion: str = None, timeout: float = 5.0, interval: float = 1.0
    ) -> Any:
        """Wait for email matching `criterion` to arrive into mailbox.

        :param criterion: email filter to wait for, defaults to ""
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0
        :return: list of messages or False

        Possible wait criterias are: SUBJECT, SENDER and BODY

        Example:

        .. code-block:: robotframework

            Wait for Email     SUBJECT:rpa task calling    timeout=300    interval=10
        """
        if self.app is None:
            raise ValueError("Requires active Outlook Application")
        if criterion is None:
            self.logger.warning(
                "Wait for message requires criteria for which message to wait for."
            )
            return False
        end_time = time.time() + float(timeout)
        namespace = self.app.GetNamespace("MAPI")
        inbox = namespace.GetDefaultFolder(6)
        while time.time() < end_time:
            messages = inbox.Items
            messages.Sort("[ReceivedTime]", True)
            for msg in messages:
                if type(msg).__name__ != "_MailItem":
                    continue
                m = self._mail_item_to_dict(msg)
                if self._is_email_too_old(m):
                    break
                else:
                    match_found = self._check_for_matching(criterion, m)
                    if match_found:
                        return match_found
            time.sleep(interval)
        raise AssertionError("Did not find matching message in the Outlook inbox")

    def get_emails(
        self,
        account_name: str = None,
        folder_name: str = None,
        email_filter: str = None,
        save_attachments: bool = False,
        attachment_folder: str = None,
        sort: bool = False,
        sort_key: str = None,
        sort_descending: bool = True,
    ) -> List:
        """Get emails from a specified email folder. Can be used to save attachments.

        :param account_name: needs to be given if there are shared accounts in use,
         defaults to None
        :param folder_name: target folder where to get emails from, default Inbox
        :param email_filter: how to filter email, default no filter,
         ie. all emails in folder
        :param save_attachments: if attachments should be saved, defaults to False
        :param attachment_folder: target folder where attachments are saved,
         defaults to current directory
        :param sort: if emails should be sorted, defaults to False
        :param sort_key: needs to be given if emails are to be sorted
        :param sort_descending: set to False for ascending sort, defaults to True
        :return: list of emails (list of dictionaries)

        Example:

        .. code-block:: robotframework

            ${emails}=  Get Emails
            ...    email_folder=priority
            ...    email_filter=[Subject]='incoming order'
            ...    save_attachments=True
            ...    attachment_folder=%{ROBOT_ROOT}${/}attachments
            ...    sort=True
            ...    sort_key=Received
            ...    sort_descending=False
        """
        folder = self._get_folder(account_name, folder_name)
        folder_messages = folder.Items if folder else []
        messages = []
        if folder_messages and email_filter:
            try:
                folder_messages = folder_messages.Restrict(email_filter)
            except Exception:
                raise AttributeError(  # pylint: disable=raise-missing-from
                    "Invalid email filter '%s'" % email_filter
                )
        if folder_messages and sort:
            sort_key = sort_key or "ReceivedTime"
            try:
                folder_messages.Sort(f"[{sort_key}]", sort_descending)
            except Exception:
                raise AttributeError(  # pylint: disable=raise-missing-from
                    "Invalid email sort key '%s'" % sort_key
                )
        for message in folder_messages:
            if type(message).__name__ != "_MailItem":
                continue
            if save_attachments:
                self.save_email_attachments(message.Attachments, attachment_folder)
            messages.append(self._mail_item_to_dict(message))
        return messages

    def _get_folder(self, account_name, email_folder):
        namespace = self.app.GetNamespace("MAPI")
        if not account_name and not email_folder:
            self.logger.warning("Getting items from default account inbox")
            return namespace.GetDefaultFolder(6)
        email_folder = email_folder or "Inbox"
        folder = None
        if account_name:
            account_folder = self._get_account_folder(namespace, account_name)
            if account_folder:
                folder = self._get_matching_folder(email_folder, account_folder)
            else:
                raise AttributeError("Did not find account by name '%s'" % account_name)
        else:
            folder = self._get_matching_folder(email_folder, None)
        return folder if folder else []

    def _get_account_folder(self, namespace, account_name):
        account_folder = None
        for f in namespace.Folders:
            if f.Name == account_name:
                account_folder = f
                break
        return account_folder

    def _get_matching_folder(self, folder_name, folder=None):
        folders = []
        if not folder or isinstance(folder, str):
            folders = self.app.GetNamespace("MAPI").Folders
        elif isinstance(folder, list):
            folders = folder
        else:
            folders = folder.Folders
        for f in folders:
            if folder_name == f.Name:
                self.logger.debug("Found matching folder: %s", f.Name)
                return f
            emails = self._get_matching_folder(folder_name, f)
            if emails:
                return emails
        return None

    def save_email_attachments(
        self, attachments: Any, attachment_folder: str, overwrite: bool = False
    ) -> None:
        """Save email attachments.

        Note. Keyword "Get Emails" can be also used to save attachments.

        :param attachments: all attachments from email or single attachment
        :param attachment_folder: target folder where attachments are saved,
            defaults to current directory
        :param overwrite: overwrite existing file if True, defaults to False

        Example:

        .. code-block:: robotframework

            ${emails} =  Get Emails
            ...    email_folder=priority
            FOR  ${email}  IN   @{emails}
                FOR  ${attachment}  IN  @{email}[Attachments]
                    IF  ${attachment}[size] < 100000   # bytes
                        Save Email Attachments
                        ...  ${attachment}
                        ...  ${CURDIR}${/}attachments
                    ELSE IF  ".pdf" in "${attachment}[filename]"
                        Save Email Attachments
                        ...  ${attachment}
                        ...  ${CURDIR}${/}attachments${/}pdf
                    END
                END
            END
        """
        attachment_target = Path(attachment_folder) if attachment_folder else Path(".")
        if isinstance(attachments, dict):
            email_attachments = [attachments["item"]]
        else:
            email_attachments = attachments
        for attachment in email_attachments:
            file_path = (attachment_target / attachment.FileName).absolute()
            if not overwrite:
                file_path = counter_duplicate_path(file_path)
            attachment.SaveAsFile(file_path)

    def mark_email_as_read(self, email: Any, read: bool = True) -> None:
        """Mark email 'read' property. Can be used to mark email as unread.

        :param email: target email
        :param read: True marks email as Read, False as Unread

        Example:

        .. code-block:: robotframework

            ${emails}=  Get Emails
            # Mark all as read
            FOR  ${email}  IN  @{emails}
                Mark Email As Read  ${email}
            END

            # Mark all as unread
            FOR  ${email}  IN  @{emails}
                Mark Email As Read  ${email}  False
            END
        """
        read_value = not read
        if type(email).__name__ == "_MailItem":
            email.UnRead = read_value
            email.Save()
        elif isinstance(email, dict):
            email["object"].UnRead = read_value
            email["object"].Save()

    def move_emails(
        self,
        account_name: str = None,
        source_folder: str = None,
        email_filter: str = None,
        target_folder: str = None,
    ) -> bool:
        """Move emails from source folder to target folder.

        Use of "account_name" is recommended if there are shared accounts in use.

        :param account_name: needs to be given if there are shared accounts in use,
         defaults to None
        :param source_folder: folder where source emails exist
        :param email_filter: how to filter email, default no filter,
         ie. all emails in folder
        :param target_folder: folder where emails are moved into
        :return: True if move operation was success, False if not

        Example:

        .. code-block:: robotframework

            # moving messages from Inbox to target_folder
            Move Emails
            ...    target_folder=Processed Invoices
            ...    email_filter=[Subject]='incoming invoice'

            # moving messages from source_folder to target_folder
            Move Emails
            ...    source_folder=Incoming Invoices
            ...    target_folder=Processed Invoices
            ...    email_filter=[Subject]='incoming invoice'
        """
        if not target_folder:
            raise AttributeError("Can't move emails without target_folder")
        folder = self._get_folder(account_name, source_folder)
        folder_messages = folder.Items if folder else []
        if folder_messages and email_filter:
            try:
                folder_messages = folder_messages.Restrict(email_filter)
            except Exception:
                raise AttributeError(  # pylint: disable=raise-missing-from
                    "Invalid email filter '%s'" % email_filter
                )
        if not folder_messages:
            self.logger.warning("Did not find emails to move")
            return False
        tf = self._get_folder(account_name, target_folder)
        if not tf or tf.Name != target_folder:
            self.logger.warning("Did not find target folder")
            return False
        self.logger.info("Found %d emails to move", len(folder_messages))
        self.logger.info("Found target folder: %s", tf.Name)
        for m in folder_messages:
            m.UnRead = False
            m.Move(tf)
            m.Save()
        return True

    def _mail_item_to_dict(self, mail_item):
        mi = mail_item
        response = {
            "Sender": self._get_sender_email_address(mi),
            "To": [],
            "CC": [],
            "BCC": [],
            "Subject": mi.Subject,
            "Body": mi.Body,
            "Attachments": [
                {"filename": a.FileName, "size": a.Size, "item": a}
                for a in mi.Attachments
            ],
            "Size": mi.Size,
            "object": mi,
        }
        rt = getattr(mail_item, "ReceivedTime", "<UNKNOWN>")
        response["ReceivedTime"] = rt.isoformat() if rt != "<UNKNOWN>" else rt
        response["ReceivedTimestamp"] = (
            datetime(
                rt.year, rt.month, rt.day, rt.hour, rt.minute, rt.second
            ).timestamp()
            if rt
            else None
        )
        so = getattr(mail_item, "SentOn", "<UNKNOWN>")
        response["SentOn"] = so.isoformat() if so != "<UNKNOWN>" else so
        self._handle_recipients(mi.Recipients, response)
        return response

    def _get_recipient_email_address(self, recipient):
        email_address = None
        try:
            email_address = recipient.AddressEntry.GetExchangeUser().PrimarySmtpAddress
        except Exception:  # pylint: disable=broad-except
            email_address = recipient.Address
        return email_address

    def _handle_recipients(self, recipients, response):
        for r in recipients:
            if r.Type == 1:
                response["To"].append(
                    {
                        "name": r.Name,
                        "email": self._get_recipient_email_address(r),
                    }
                )
            elif r.Type == 2:
                response["CC"].append(
                    {
                        "name": r.Name,
                        "email": self._get_recipient_email_address(r),
                    }
                )
            elif r.Type == 3:
                response["BCC"].append(
                    {
                        "name": r.Name,
                        "email": self._get_recipient_email_address(r),
                    }
                )

    def _get_sender_email_address(self, mail_item):
        mi = mail_item
        return (
            mi.Sender.GetExchangeUser().PrimarySmtpAddress
            if mi.SenderEmailType == "EX"
            else mi.SenderEmailAddress
        )

    def _address_entry_to_dict(self, address_entry):
        ae = address_entry
        return {"Name": ae.Name, "Address": ae.Address}



import datetime
import email
import logging
import re
import time
from email import policy  # pylint: disable=no-name-in-module
from enum import Enum
from multiprocessing import AuthenticationError
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import pytz
from exchangelib import (
    DELEGATE,
    IMPERSONATION,
    OAUTH2,
    UTC,
    Account,
    Configuration,
    Credentials,
    FileAttachment,
    Folder,
    HTMLBody,
    Identity,
    ItemAttachment,
    Mailbox,
    Message,
    OAuth2AuthorizationCodeCredentials,
)
from exchangelib.folders import Inbox
from oauthlib.oauth2 import OAuth2Token

from RPA.Email.common import (
    OAuthConfig,
    OAuthMixin,
    OAuthProvider,
    counter_duplicate_path,
)
from RPA.Robocorp.Vault import Vault
from RPA.Robocorp.utils import protect_keywords


EMAIL_CRITERIA_KEYS = {
    "subject": "subject",
    "subject_contains": "subject__contains",
    "body": "body",
    "body_contains": "body__contains",
    "sender": "sender",
    "sender_contains": "sender__contains",
    "before": "datetime_received__range",
    "after": "datetime_received__range",
    "between": "datetime_received__range",
    "category": "categories",
    "category_contains": "categories__contains",
    "importance": "importance",
}

lib_vault = Vault()


def mailbox_to_email_address(mailbox):
    return {
        "name": mailbox.name if hasattr(mailbox, "name") else "",
        "email_address": mailbox.email_address
        if hasattr(mailbox, "email_address")
        else "",
    }


class AccessType(Enum):
    """Authorization access type."""

    DELEGATE = DELEGATE
    IMPERSONATION = IMPERSONATION


class NoRecipientsError(ValueError):
    """Raised when email to be sent does not have any recipients, cc or bcc addresses."""  # noqa: E501


class OAuth2Creds(OAuth2AuthorizationCodeCredentials):

    """OAuth2 auth code flow credentials wrapper supporting token state on refresh."""

    def __init__(
        self, *args, on_token_refresh: Callable, oauth_provider: OAuthConfig, **kwargs
    ):
        super().__init__(*args, **kwargs)
        # Additional behaviour to trigger during token refresh (when it expires), like
        #  saving the newly generated structure in the Vault.
        self._on_token_refresh = on_token_refresh
        self._token_url = oauth_provider.token_url
        self._scope = oauth_provider.scope

    def on_token_auto_refreshed(self, access_token: OAuth2Token):
        """Saves the newly obtained token internally and in the Vault."""
        super().on_token_auto_refreshed(access_token)
        self._on_token_refresh(access_token)

    @property
    def token_url(self) -> str:
        """Custom token URL coming from the OAuth2 provider settings."""
        return self._token_url

    @property
    def scope(self) -> List[str]:
        """Custom permissions list coming from the OAuth2 provider settings."""
        return self._scope.split()


class Exchange(OAuthMixin):
    """`Exchange` is a library for sending, reading, and deleting emails.
    `Exchange` is interfacing with Exchange Web Services (EWS).

    For more information about server settings, see
    `this Microsoft support article <https://support.microsoft.com/en-us/office/server-settings-you-ll-need-from-your-email-provider-c82de912-adcc-4787-8283-45a1161f3cc3>`_.

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Email.Exchange
        ...     vault_name=email_oauth_microsoft    vault_token_key=token
        ...     tenant=ztzvn.onmicrosoft.com  # your custom tenant here
        Task Setup      Ensure Auth

        *** Variables ***
        ${ACCOUNT}              ACCOUNT_NAME
        ${RECIPIENT_ADDRESS}    RECIPIENT
        ${IMAGES}               myimage.png
        ${ATTACHMENTS}          C:${/}files${/}mydocument.pdf

        *** Keywords ***
        Ensure Auth
            ${secrets} =    Get Secret    email_oauth_microsoft
            RPA.Email.Exchange.Authorize    ${ACCOUNT}
            ...    is_oauth=${True}  # use the OAuth2 auth code flow (required)
            ...    client_id=${secrets}[client_id]  # app ID
            ...    client_secret=${secrets}[client_secret]  # app password
            ...    token=${secrets}[token]  # token dict (access, refresh, scope etc.)

        *** Tasks ***
        Task of sending email
            Send Message  recipients=${RECIPIENT_ADDRESS}
            ...           subject=Exchange Message from RPA Robot
            ...           body=<p>Exchange RPA Robot message body<br><img src='myimage.png'/></p>
            ...           save=${TRUE}
            ...           html=${TRUE}
            ...           images=${IMAGES}
            ...           cc=EMAIL_ADDRESS
            ...           bcc=EMAIL_ADDRESS
            ...           attachments=${ATTACHMENTS}

        Task of listing messages
            # Attachments are saved specifically with a keyword Save Attachments
            ${messages}=    List Messages
            FOR    ${msg}    IN    @{messages}
                Log Many    ${msg}
                ${attachments}=    Run Keyword If    "${msg}[subject]"=="about my orders"
                ...    Save Attachments
                ...    ${msg}
                ...    save_dir=${CURDIR}${/}savedir
            END
            # Using save_dir all attachments in listed messages are saved
            ${messages}=    List Messages
            ...    INBOX/Problems/sub1
            ...    criterion=subject:'about my orders'
            ...    save_dir=${CURDIR}${/}savedir2
            FOR    ${msg}    IN    @{messages}
                Log Many    ${msg}
            END

        Task of moving messages
            Move Messages    criterion=subject:'about my orders'
            ...    source=INBOX/Processed Purchase Invoices/sub2
            ...    target=INBOX/Problems/sub1

    **Python**

    .. code-block:: python

        from RPA.Email.Exchange import Exchange
        from RPA.Robocorp.Vault import Vault

        vault_name = "email_oauth_microsoft"
        secrets = Vault().get_secret(vault_name)
        ex_account = "ACCOUNT_NAME"

        mail = Exchange(
            vault_name=vault_name,
            vault_token_key="token",
            tenant="ztzvn.onmicrosoft.com"
        )
        mail.authorize(
            username=ex_account,
            is_oauth=True,
            client_id=secrets["client_id"],
            client_secret=secrets["client_secret"],
            token=secrets["token"]
        )
        mail.send_message(
            recipients="RECIPIENT",
            subject="Message from RPA Python",
            body="RPA Python message body",
        )

    **OAuth2**

    The OAuth2 flow is the only way of authorizing at the moment as Microsoft disabled
    entirely the usage of passwords, even App Passwords. And since you have to work
    with tokens now and because this library has the capability to automatically
    refresh an expired token, please don't forget to initialize the library with the
    following parameters: `vault_name`, `vault_token_key` and `tenant`.

    Learn more on how to use the OAuth2 flow in this Portal robot
    `example-oauth-email <https://github.com/robocorp/example-oauth-email>`_.

    **About criterion parameter**

    Following table shows possible criterion keys that can be used to filter emails.
    There apply to all keywords which have ``criterion`` parameter.

    ================= ================
    Key               Effective search
    ================= ================
    subject           subject to match
    subject_contains  subject to contain
    body              body to match
    body_contains     body to contain
    sender            sender (from) to match
    sender_contains   sender (from) to contain
    before            received time before this time
    after             received time after this time
    between           received time between start and end
    category          categories to match
    category_contains categories to contain
    importance        importance to match
    ================= ================

    Keys `before`, `after` and `between` at the moment support two
    different timeformats either `%d-%m-%Y %H:%M` or `%d-%m-%Y`. These
    keys also support special string `NOW` which can be used especially
    together with keyword ``Wait for message  criterion=after:NOW``.

    When giving time which includes hours and minutes then the whole
    time string needs to be enclosed into single quotes.

    .. code-block:: bash

        before:25-02-2022
        after:NOW
        between:'31-12-2021 23:50 and 01-01-2022 00:10'

    Different criterion keys can be combined.

    .. code-block:: bash

        subject_contains:'new year' between:'31-12-2021 23:50 and 01-01-2022 00:10'

    Please **note** that all values in the criterion that contain spaces need
    to be enclosed within single quotes.

    In the following example the email `subject` is going to matched
    only against `new` not `new year`.

    .. code-block:: bash

        subject_contains:new year

    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    TO_PROTECT = ["authorize"] + OAuthMixin.TO_PROTECT

    def __init__(
        self,
        vault_name: Optional[str] = None,
        vault_token_key: Optional[str] = None,
        tenant: Optional[str] = None,
    ) -> None:
        # Init the OAuth2 support. (mandatory usage)
        self._tenant = tenant or "common"
        super().__init__(OAuthProvider.MICROSOFT, tenant=self._tenant)

        protect_keywords("RPA.Email.Exchange", self.TO_PROTECT)
        self.logger = logging.getLogger(__name__)

        self._vault_name = vault_name
        self._vault_token_key = vault_token_key

        self.credentials = None
        self.config = None
        self.account = None
        self._saved_attachments = []

    def _on_token_refresh(self, token: OAuth2Token):
        """Callable you can override in order to save the newly obtained token in a
        safe place.
        """
        if self._vault_name and self._vault_token_key:
            token = dict(token)
            self._sync_token_metadata(token)
            secret = lib_vault.get_secret(self._vault_name)
            secret[self._vault_token_key] = token
            lib_vault.set_secret(secret)
            self.logger.info(
                "OAuth2 token was refreshed in Vault %r as %r. (new expiry: %d)",
                self._vault_name,
                self._vault_token_key,
                token["expires_at"],
            )
        else:
            self.logger.warning(
                "OAuth2 token was refreshed but didn't get saved in Vault as well."
                " (import the library with `vault_name` and `vault_token_key` set in"
                " order to avoid this)"
            )

    def authorize(
        self,
        username: str,
        password: Optional[str] = None,
        autodiscover: bool = True,
        access_type: Union[AccessType, str] = AccessType.DELEGATE,
        server: Optional[str] = None,
        primary_smtp_address: Optional[str] = None,
        is_oauth: bool = False,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[dict] = None,
    ) -> None:
        """Connect to Exchange account

        :param username: account username
        :param password: account password (can be skipped with OAuth2)
        :param autodiscover: use autodiscover or set it off (on by default)
        :param access_type: default "DELEGATE", other option "IMPERSONATION"
        :param server: required for configuration setting (with `autodiscover` off)
        :param primary_smtp_address: by default set to username, but can be
            set to be different from username
        :param is_oauth: use the OAuth2 authorization code flow (instead of basic auth)
        :param client_id: registered application ID
        :param client_secret: registered application secret (password)
        :param token: contains access and refresh tokens, type, scope, expiry etc.
        """
        kwargs = {}
        kwargs["autodiscover"] = autodiscover
        access_type = (
            access_type
            if isinstance(access_type, AccessType)
            else AccessType(access_type.lower())
        )
        kwargs["access_type"] = access_type.value
        kwargs["primary_smtp_address"] = (
            primary_smtp_address if primary_smtp_address else username
        )
        if is_oauth:
            self.credentials = OAuth2Creds(
                on_token_refresh=self._on_token_refresh,
                oauth_provider=self._oauth_provider,
                tenant_id=self._tenant,
                identity=Identity(upn=username),
                client_id=client_id,
                client_secret=client_secret,
                # Contains at least a non-expired access token or non-revoked refresh
                #  one. (otherwise an authorization code should be present inside)
                access_token=OAuth2Token(params=token) if token else None,
            )
        else:
            if password is None:
                raise ValueError("A password should be provided with basic auth")
            self.credentials = Credentials(username, password)
        if server:
            self.config = Configuration(
                server=server,
                credentials=self.credentials,
                # Automatically detects authentication type based on the provided
                #  `credentials` object. (when not using OAuth2)
                auth_type=OAUTH2 if is_oauth else None,
            )
            kwargs["config"] = self.config
        else:
            kwargs["credentials"] = self.credentials

        if self.config and autodiscover:
            self.logger.warning(
                "Autodiscovery is left ON while using custom configuration, you may "
                "need to turn it OFF in order to use the custom setting."
            )
        self.account = Account(**kwargs)

    def list_messages(
        self,
        folder_name: Optional[str] = None,
        criterion: Optional[str] = None,
        contains: Optional[bool] = False,  # pylint: disable=unused-argument
        count: Optional[int] = 100,
        save_dir: Optional[str] = None,
    ) -> list:
        """List messages in the account inbox. Order by descending
        received time.

        :param folder_name: name of the email folder, default INBOX
        :param criterion: list messages matching criterion
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
        :param count: number of messages to list
        :param save_dir: set to path where attachments should be saved,
         default None (attachments are not saved)
        """
        # pylint: disable=no-member
        messages = []
        source_folder = self._get_folder_object(folder_name)
        if criterion:
            filter_dict = self._get_filter_key_value(criterion)
            items = source_folder.filter(**filter_dict)
        else:
            items = source_folder.all()
        for item in items.order_by("-datetime_received")[:count]:
            attachments = []
            if save_dir and len(item.attachments) > 0:
                # NOTE(cmin764): Default `overwrite` param is assumed here.
                attachments = self._save_attachments(item, save_dir)
            messages.append(self._get_email_details(item, attachments))
        return messages

    def list_unread_messages(
        self,
        folder_name: Optional[str] = None,
        criterion: Optional[str] = None,
        contains: Optional[bool] = False,
        count: Optional[int] = 100,
        save_dir: Optional[str] = None,
    ) -> list:
        """List unread messages in the account inbox. Order by descending
        received time.

        :param folder_name: name of the email folder, default INBOX
        :param criterion: list messages matching criterion
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
        :param count: number of messages to list
        :param save_dir: set to path where attachments should be saved,
         default None (attachments are not saved)
        """
        messages = self.list_messages(folder_name, criterion, contains, count, save_dir)
        return [m for m in messages if not m["is_read"]]

    def _get_inbox_folder(
        self, folder_name: str, parent_folder: Optional[str] = None
    ) -> Inbox:
        if parent_folder is None or parent_folder is self.account.inbox:
            parent = self.account.inbox
        else:
            parent = self.account.inbox / parent_folder
        return parent / folder_name if folder_name else parent

    def send_message(
        self,
        recipients: Optional[Union[List[str], str]] = None,
        subject: Optional[str] = "",
        body: Optional[str] = "",
        attachments: Optional[Union[List[str], str]] = None,
        html: Optional[bool] = False,
        images: Optional[Union[List[str], str]] = None,
        cc: Optional[Union[List[str], str]] = None,
        bcc: Optional[Union[List[str], str]] = None,
        save: Optional[bool] = False,
    ) -> None:
        """Keyword for sending message through connected Exchange account.

        :param recipients: list of email addresses
        :param subject: message subject, defaults to ""
        :param body: message body, defaults to ""
        :param attachments: list of filepaths to attach, defaults to `None`
        :param html: if message content is in HTML, default `False`
        :param images: list of filepaths for inline use, defaults to `None`
        :param cc: list of email addresses
        :param bcc: list of email addresses
        :param save: is sent message saved to Sent messages folder or not,
            defaults to False

        Email addresses can be prefixed with ``ex:`` to indicate an Exchange
        account address.

        At least one target needs to exist for `recipients`, `cc` or `bcc`.
        """
        if not self.account:
            raise AuthenticationError("Not authorized to any Exchange account")
        recipients, cc, bcc, attachments, images = self._handle_message_parameters(
            recipients, cc, bcc, attachments, images
        )
        if not recipients and not cc and not bcc:
            raise NoRecipientsError(
                "At least one address is required for 'recipients', 'cc' or 'bcc' parameter"  # noqa: E501
            )
        self.logger.info("Sending message to %s", ",".join(recipients))

        m = Message(
            account=self.account,
            subject=subject,
            body=body,
            to_recipients=recipients,
            cc_recipients=cc,
            bcc_recipients=bcc,
        )

        self._add_attachments_to_msg(attachments, m)
        self._add_images_inline_to_msg(images, html, body, m)

        if html:
            m.body = HTMLBody(body)
        else:
            m.body = body

        # TODO. The exchangelib does not seem to provide any straightforward way of
        # verifying if message was sent or not
        if save:
            m.folder = self.account.sent
            m.send_and_save()
        else:
            m.send()

    def _handle_message_parameters(self, recipients, cc, bcc, attachments, images):
        recipients = recipients or []
        cc = cc or []
        bcc = bcc or []
        attachments = attachments or []
        images = images or []
        if not isinstance(recipients, list):
            recipients = recipients.split(",")
        if not isinstance(cc, list):
            cc = cc.split(",")
        if not isinstance(bcc, list):
            bcc = bcc.split(",")
        if not isinstance(attachments, list):
            attachments = str(attachments).split(",")
        if not isinstance(images, list):
            images = str(images).split(",")
        recipients, cc, bcc = self._handle_recipients(recipients, cc, bcc)
        return recipients, cc, bcc, attachments, images

    def _handle_recipients(self, recipients, cc, bcc):
        recipients = [
            Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p
            for p in recipients
        ]
        cc = [Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p for p in cc]
        bcc = [
            Mailbox(email_address=p.split("ex:")[1]) if "ex:" in p else p for p in bcc
        ]
        return recipients, cc, bcc

    def _add_attachments_to_msg(self, attachments, msg):
        for attachment in attachments:
            attachment = attachment.strip()
            with open(attachment, "rb") as f:
                atname = str(Path(attachment).name)
                fileat = FileAttachment(name=atname, content=f.read())
                msg.attach(fileat)

    def _add_images_inline_to_msg(self, images, html, body, msg):
        for image in images:
            image = image.strip()
            with open(image, "rb") as f:
                imname = str(Path(image).name)
                fileat = FileAttachment(
                    name=imname, content=f.read(), content_id=imname
                )
                msg.attach(fileat)
                if html:
                    body = body.replace(imname, f"cid:{imname}")

    def create_folder(self, folder_name: str, parent_folder: Optional[str] = None):
        """Create email folder.

        :param folder_name: name for the new folder (required)
        :param parent_folder: name for the parent folder, by default INBOX
        """
        parent = self._get_inbox_folder(folder_name, parent_folder)
        self.logger.info("Create folder %r", folder_name)
        new_folder = Folder(parent=parent, name=folder_name)
        new_folder.save()

    def delete_folder(self, folder_name: str, parent_folder: Optional[str] = None):
        """Delete email folder.

        :param folder_name: current folder name (required)
        :param parent_folder: name for the parent folder, by default INBOX
        """
        folder_to_delete = self._get_inbox_folder(folder_name, parent_folder)
        self.logger.info("Delete folder %r", folder_name)
        folder_to_delete.delete()

    def rename_folder(
        self,
        oldname: str,
        newname: str,
        parent_folder: Optional[str] = None,
    ):
        """Rename email folder

        :param oldname: current folder name
        :param newname: new name for the folder
        :param parent_folder: name for the parent folder, by default INBOX
        :return: True if operation was successful, False if not
        """
        parent = self._get_inbox_folder("", parent_folder)
        self.logger.info("Rename folder %r to %r", oldname, newname)
        items = self._get_inbox_folder(oldname, parent_folder).all()
        old_folder = Folder(parent=parent, name=oldname)
        old_folder.name = newname
        old_folder.save()
        items.move(to_folder=parent / newname)
        self.delete_folder(oldname, parent_folder)

    def empty_folder(
        self,
        folder_name: str,
        parent_folder: Optional[str] = None,
        delete_sub_folders: Optional[bool] = False,
    ):
        """Empty email folder of all items

        :param folder_name: current folder name (required)
        :param parent_folder: name for the parent folder, by default INBOX
        :param delete_sub_folders: delete sub folders or not, by default False
        :return: True if operation was successful, False if not
        """
        if parent_folder is None:
            empty_folder = self._get_folder_object(folder_name)
        else:
            empty_folder = self._get_folder_object(f"{parent_folder} / {folder_name}")
        self.logger.info("Empty folder %r", empty_folder)
        empty_folder.empty(delete_sub_folders=delete_sub_folders)

    def move_messages(
        self,
        criterion: Optional[str] = "",
        source: Optional[str] = None,
        target: Optional[str] = None,
        contains: Optional[bool] = False,  # pylint: disable=unused-argument
    ) -> bool:
        """Move message(s) from source folder to target folder

        :param criterion: move messages matching this criterion
        :param source: source folder
        :param target: target folder
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
        :return: boolean result of operation, True if 1+ items were moved else False

        Criterion examples:

        - subject:my message subject
        - body:something in body
        - sender:sender@domain.com
        """
        source_folder = self._get_folder_object(source)
        target_folder = self._get_folder_object(target)
        if source_folder == target_folder:
            raise KeyError("Source folder is same as target folder")
        filter_dict = self._get_filter_key_value(criterion)
        items = source_folder.filter(**filter_dict)
        if items and items.count() > 0:
            items.move(to_folder=target_folder)
            return True
        else:
            self.logger.warning("No items match criterion '%s'", criterion)
            return False

    def move_message(
        self,
        msg: Optional[dict],
        target: Optional[str],
    ):
        """Move a message into target folder

        :param msg: dictionary of the message
        :param target: path to target folder
        :raises AttributeError: if `msg` is not a dictionary containing
         `id` and `changekey` attributes

        Example:

        .. code-block:: robotframework

            ${messages}=    List Messages
            ...    INBOX
            ...    criterion=subject:about my orders
            FOR    ${msg}    IN    @{messages}
                Run Keyword If    "${msg}[sender][email_address]"=="${priority_account}"
                ...    Move Message    ${msg}    target=INBOX / Problems / priority
            END
        """
        if not all(k in msg for k in ["id", "changekey"]):
            raise AttributeError(
                "Move Message keyword expects message dictionary "
                'containing "id" and "changekey" attributes'
            )
        message_id = [(msg["id"], msg["changekey"])]
        target_folder = self._get_folder_object(target)
        self.account.bulk_move(ids=message_id, to_folder=target_folder)

    def _get_folder_object(self, folder_name):
        if not folder_name:
            return self.account.inbox

        folders = folder_name.split("/")
        if "inbox" in folders[0].lower():
            folders[0] = self.account.inbox
        folder_object = None
        for folder in folders:
            if folder_object:
                folder_object /= folder.strip()
            else:
                folder_object = folder

        return folder_object

    def _get_filter_key_value(self, criterion):
        regex1 = rf"({':|'.join(EMAIL_CRITERIA_KEYS)}:|or|and)'(.*?)'"
        regex2 = rf"({':|'.join(EMAIL_CRITERIA_KEYS)}:|or|and)(\S*)\s*"
        parts = re.findall(regex1, criterion, re.IGNORECASE)
        valid_filters = {}
        for part in parts:
            res = self._parse_email_criteria(part)
            if not res or len(res) != 2:
                continue
            self.logger.debug("First regex pass: %s %s", res[0], res[1])
            if res[0] not in valid_filters.keys():
                valid_filters[res[0]] = res[1]
        parts = re.findall(regex2, criterion, re.IGNORECASE)
        for part in parts:
            res = self._parse_email_criteria(part)
            if not res or len(res) != 2:
                continue
            self.logger.debug("Second regex pass: %s %s", res[0], res[1])
            if res[0] not in valid_filters.keys():
                valid_filters[res[0]] = res[1]
        if criterion and criterion != "" and len(valid_filters) == 0:
            raise KeyError("Invalid criterion '%s'" % criterion)
        self.logger.info(
            "Using filter: %s",
            ",".join(["{}:{}".format(k, v) for k, v in valid_filters.items()]),
        )
        return valid_filters

    def _parse_email_criteria(self, part):
        original_key, value = part
        original_key = original_key.replace(":", "")
        if original_key in ["and", "or", "contains"]:
            # Note. Not implemented yet
            return None
        key = None
        if original_key in EMAIL_CRITERIA_KEYS.keys():
            key = EMAIL_CRITERIA_KEYS[original_key]
        else:
            raise KeyError("Unknown email criteria key '%s'" % original_key)

        if value.startswith("'") and value.endswith("'"):
            value = value[1:-1]

        value = self._handle_specific_keys(original_key, value)
        self.logger.debug("Returning parsed criteria (%s, %s)", key, value)
        return key, value

    def _parse_date_from_string(self, date_string):
        date = None
        if date_string.upper() == "NOW":
            right_now = datetime.datetime.now()
            return right_now.astimezone(pytz.utc)
        try:
            date = datetime.datetime.strptime(date_string, "%d-%m-%Y %H:%M")
            return self._date_for_exchange(date)
        except ValueError:
            pass
        try:
            date = datetime.datetime.strptime(date_string, "%d-%m-%Y")
            return self._date_for_exchange(date)
        except ValueError:
            pass
        return date

    def _date_for_exchange(self, date):
        return datetime.datetime(
            date.year,
            date.month,
            date.day,
            date.hour,
            date.minute,
            tzinfo=UTC,
        )

    def _handle_specific_keys(self, key, value):
        if key.upper() == "BEFORE":
            start = datetime.datetime(1972, 1, 1, tzinfo=UTC)
            end = self._parse_date_from_string(value)
            value = (start, end)
        if key.upper() == "AFTER":
            start = self._parse_date_from_string(value)
            end = datetime.datetime(2050, 1, 1, tzinfo=UTC)
            value = (start, end)
        if key.upper() == "BETWEEN":
            ranges = value.upper().split(" AND ")
            if len(ranges) == 2:
                start = self._parse_date_from_string(ranges[0])
                end = self._parse_date_from_string(ranges[1])
                value = (start, end)
            else:
                return None
        if key.upper() == "IMPORTANCE":
            try:
                value = value.capitalize()
            except ValueError:
                return None
        return value

    def wait_for_message(
        self,
        criterion: Optional[str] = "",
        timeout: Optional[float] = 5.0,
        interval: Optional[float] = 1.0,
        contains: Optional[bool] = False,  # pylint: disable=unused-argument
        save_dir: Optional[str] = None,
    ) -> Any:
        """Wait for email matching `criterion` to arrive into INBOX.

        :param criterion: wait for message matching criterion
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0 (minimum)
        :param contains: if matching should be done using `contains` matching
         and not `equals` matching, default `False` is means `equals` matching
         THIS PARAMETER IS DEPRECATED AS OF rpaframework 12.9.0
        :param save_dir: set to path where attachments should be saved,
         default None (attachments are not saved)
        :return: list of messages
        """
        self.logger.info("Wait for messages")
        end_time = time.time() + float(timeout)
        filter_dict = self._get_filter_key_value(criterion)
        items = None
        # minimum interval is 1.0 seconds
        interval = max(interval, 1.0)
        while time.time() < end_time:
            items = self.account.inbox.filter(**filter_dict)  # pylint: disable=E1101
            if items.count() > 0:
                break
            time.sleep(interval)
        messages = []
        for item in items:
            attachments = []
            if save_dir and len(item.attachments) > 0:
                # NOTE(cmin764): Default `overwrite` param is assumed here.
                attachments = self._save_attachments(item, save_dir)
            messages.append(self._get_email_details(item, attachments))

        if len(messages) == 0:
            self.logger.info("Did not receive any matching items")
        return messages

    def _save_attachments(
        self,
        item,
        save_dir: str,
        attachments_from_emls: bool = False,
        overwrite: bool = False,
    ):
        self._saved_attachments = self._saved_attachments or []
        incoming_items = item.attachments if hasattr(item, "attachments") else item

        for attachment in incoming_items:
            self.logger.debug("Attachment type: %s", type(attachment))
            local_path = Path(save_dir) / attachment.name
            if not overwrite:
                local_path = counter_duplicate_path(local_path)

            if isinstance(attachment, FileAttachment):
                with open(local_path, "wb") as stream_out, attachment.fp as stream_in:
                    buffer = stream_in.read(1024)
                    while buffer:
                        stream_out.write(buffer)
                        buffer = stream_in.read(1024)
            elif isinstance(attachment, ItemAttachment):
                with open(local_path, "wb") as message_out:
                    message_out.write(attachment.item.mime_content)
                if attachments_from_emls:
                    self._save_attachments(
                        attachment.item,
                        save_dir,
                        attachments_from_emls=False,
                        overwrite=overwrite,
                    )
            else:
                self.logger.warning(
                    "Unrecognized attachment type: %s", type(attachment)
                )
                return self._saved_attachments

            self.logger.info("Attachment saved to: %s", local_path)
            self._saved_attachments.append(
                self._new_attachment_dictionary(attachment, local_path)
            )

        return self._saved_attachments

    @staticmethod
    def _new_attachment_dictionary(
        attachment: Union[FileAttachment, ItemAttachment], local_path: Path
    ) -> Dict:
        return {
            "name": attachment.name,
            "content_type": attachment.content_type,
            "size": attachment.size,
            "is_contact_photo": getattr(attachment, "is_contact_photo", False),
            "local_path": str(local_path),
        }

    def _get_email_details(self, item, attachments):
        return {
            "subject": item.subject,
            "sender": mailbox_to_email_address(item.sender),
            "datetime_received": item.datetime_received,
            "folder": str(self._get_folder_object(item.folder)),
            "body": item.body,
            "text_body": item.text_body,
            "received_by": mailbox_to_email_address(item.received_by),
            "cc_recipients": [mailbox_to_email_address(cc) for cc in item.cc_recipients]
            if item.cc_recipients
            else [],
            "bcc_recipients": [
                mailbox_to_email_address(bcc) for bcc in item.bcc_recipients
            ]
            if item.bcc_recipients
            else [],
            "is_read": item.is_read,
            "importance": item.importance,
            "message_id": item.message_id,
            "size": item.size,
            "categories": item.categories,
            "has_attachments": len(item.attachments) > 0,
            "attachments": attachments,
            "attachments_object": item.attachments,
            "id": item.id,
            "changekey": item.changekey,
            "mime_content": item.mime_content,
        }

    def save_attachments(
        self,
        message: Union[dict, str],
        save_dir: Optional[str] = None,
        attachments_from_emls: bool = False,
        overwrite: bool = False,
    ) -> list:
        """Save attachments from message into given directory.

        :param message: dictionary or .eml file path containing message details
        :param save_dir: file path where attachments will be saved
        :param attachments_from_emls: pass `True` if the attachment is an EML file (for
            saving attachments from that EML file instead), `False` otherwise (default)
        :param overwrite: overwrite existing downloaded attachments with the same name
            if set to `True`, `False` otherwise (default)
        :return: list of saved attachments

        Example:

        .. code:: robotframework

            ${messages} =    List Messages
            FOR    ${msg}    IN    @{messages}
                Save Attachments    ${msg}    %{ROBOT_ARTIFACTS}
                ...    attachments_from_emls=${True}
            END

            ${attachments} =    Save Attachments    ${CURDIR}${/}saved.eml
            ...    %{ROBOT_ARTIFACTS}    overwrite=${True}
        """  # noqa: E501
        self._saved_attachments = []
        if isinstance(message, dict):
            return self._save_attachments(
                message["attachments_object"],
                save_dir,
                attachments_from_emls=attachments_from_emls,
                overwrite=overwrite,
            )
        else:
            # extract attachments from .eml file
            absolute_filepath = Path(message).resolve()
            if absolute_filepath.suffix != ".eml":
                raise ValueError("Filename extension needs to be '.eml'")
            return self._save_attachments_from_file(
                message, save_dir, overwrite=overwrite
            )

    def _save_attachments_from_file(
        self, filename: str, save_dir: str, *, overwrite: bool
    ):
        """
        Try to extract the attachments from given .eml file
        """
        # ensure that an output dir exists
        attachments = []
        with open(filename, "r") as f:  # pylint: disable=unspecified-encoding
            msg = email.message_from_file(  # pylint: disable=no-member
                f, policy=policy.default
            )
            for attachment in msg.iter_attachments():
                output_filename = attachment.get_filename()
                # If no attachments are found, skip this file
                if output_filename:
                    local_path = Path(save_dir) / output_filename
                    if not overwrite:
                        local_path = counter_duplicate_path(local_path)
                    with open(local_path, "wb") as of:
                        payload = attachment.get_payload(decode=True)
                        of.write(payload)
                        attachments.append(
                            {
                                "name": output_filename,
                                "content_type": None,
                                "size": len(payload),
                                "is_contact_photo": None,
                                "local_path": str(local_path),
                            }
                        )
            if len(attachments) == 0:
                self.logger.warning("No attachment found for file %s!", f.name)
        return attachments

    def save_message(self, message: dict, filename: str):
        """Save email as .eml file.

        :param message: dictionary containing message details
        :param filename: name of the file to save message into
        """
        absolute_filepath = Path(filename).resolve()
        if absolute_filepath.suffix != ".eml":
            raise ValueError("Filename extension needs to be '.eml'")
        with open(absolute_filepath, "wb") as message_out:
            message_out.write(message["mime_content"])



"""Testing of RPA.DocumentAI library."""


import pytest
from unittest import mock

from RPA.DocumentAI import DocumentAI
from RPA.DocumentAI.Base64AI import Base64AI
from RPA.DocumentAI.DocumentAI import EngineName
from RPA.DocumentAI.Nanonets import Nanonets

from . import RESOURCES_DIR


class TestDocumentAI:
    """Tests `RPA.DocumentAI`."""

    @staticmethod
    @pytest.fixture
    def library():
        return DocumentAI()

    @staticmethod
    @pytest.fixture
    def invoice_png():
        return RESOURCES_DIR / "invoice.png"

    def test_init_engine_google(self, library):
        # As `rpaframework-google` isn't installed by default.
        with pytest.raises(ImportError):
            library.init_engine("google")

    def test_init_engine_base64(self, library):
        secret = "cosmin@robocorp.com,api-key"
        with pytest.raises(TypeError):
            # Two arguments (positional or keyword) are required.
            library.init_engine("base64ai", secret="api-key")
        library.init_engine("base64ai", secret=secret)
        engine = library.engine
        assert isinstance(engine, Base64AI)
        assert secret.replace(",", ":") in engine._request_headers["Authorization"]

    def test_init_engine_nanonets(self, library):
        secret = "api-key"
        library.init_engine("nanonets", secret=secret)
        engine = library.engine
        assert isinstance(engine, Nanonets)
        assert secret == library.engine.apikey

    def test_switch_engine(self, library):
        with pytest.raises(RuntimeError):
            assert library.engine  # no engine initialized

        library.init_engine("base64ai", secret="cosmin@robocorp.com,api-key")
        library.init_engine("nanonets", secret="api-key")

        assert isinstance(library.engine, Nanonets)
        library.switch_engine(EngineName.BASE64)
        assert isinstance(library.engine, Base64AI)

    @mock.patch("RPA.DocumentAI.Base64AI.requests")
    def test_predict_base64(self, mock_requests, library, invoice_png):
        library.init_engine("base64ai", secret="cosmin@robocorp.com,api-key")
        model = "invoice-model"
        library.predict(invoice_png, model=model)
        args, kwargs = mock_requests.request.call_args
        assert args == ("POST", "https://base64.ai/api/scan")
        assert model in kwargs["data"]

    @mock.patch("RPA.DocumentAI.Nanonets.requests")
    def test_predict_nanonets(self, mock_requests, library, invoice_png):
        library.init_engine("nanonets", secret="api-key")
        model = "invoice-model"
        library.predict(invoice_png, model=model)
        args, kwargs = mock_requests.post.call_args
        assert args == (
            "https://app.nanonets.com/api/v2/OCR/Model/invoice-model/LabelFile/",
        )
        assert kwargs["files"]["file"].name == str(invoice_png)

    @pytest.mark.parametrize("extended", [False, True])
    def test_get_result(self, library, extended):
        library.init_engine("nanonets", secret="api-key")
        prediction = {
            "result": [
                {
                    "prediction": [
                        {
                            "id": "251fed51-e1fa-477e-af9d-3fe115486290",
                            "label": "bank_name",
                            "xmin": 901,
                            "ymin": 1619,
                            "xmax": 1001,
                            "ymax": 1629,
                            "score": 0.5428322,
                            "ocr_text": "Deutsche Bank",
                            "type": "field",
                            "status": "correctly_predicted",
                            "page_no": 0,
                            "label_id": "d5513122-0a46-4927-b923-cd9e6eed758d",
                        },
                        {"extra": "field"},
                    ]
                }
            ]
        }
        library._results[library._active_engine] = prediction
        result = library.get_result(extended=extended)
        if extended:
            assert result == prediction
        else:
            assert result[0]["ocr_text"] == "Deutsche Bank"
